from __future__ import annotations

import functools
import json
import sys
from collections.abc import Callable
from importlib import resources
from pathlib import Path
from typing import Any

histogram_file = resources.files("uhi") / "resources/histogram.schema.json"

__all__ = ["histogram_file", "validate"]


def __dir__() -> list[str]:
    return __all__


@functools.cache
def _histogram_schema() -> Callable[[dict[str, Any]], None]:
    import fastjsonschema  # noqa: PLC0415

    with histogram_file.open(encoding="utf-8") as f:
        return fastjsonschema.compile(json.load(f))  # type: ignore[no-any-return]


def validate(data: dict[str, Any]) -> None:
    """Validate a histogram object against the schema."""
    validate = _histogram_schema()
    validate(data)


def main(*files: str) -> None:
    """Validate histogram files."""
    import fastjsonschema  # noqa: PLC0415

    retval = 0

    for file in files:
        with Path(file).open(encoding="utf-8") as f:
            data = json.load(f)
        try:
            validate(data)
        except fastjsonschema.JsonSchemaValueException as e:
            print(f"ERROR {file}: {e.message}")  # noqa: T201
            retval = 1
        else:
            print(f"OK {file}")  # noqa: T201

    if retval:
        raise SystemExit(retval)


if __name__ == "__main__":
    main(*sys.argv[1:])



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from __future__ import annotations

import importlib.metadata
import os

# -- Project information -----------------------------------------------------

project = "uhi"
copyright = "2021, Henry Schreiner, Hans Dembinski, Jim Pivarski"
author = "Henry Schreiner, Hans Dembinski, Jim Pivarski"

# The full version, including alpha/beta/rc tags
version = release = importlib.metadata.version("uhi")


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "sphinx-jsonschema",
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
    "sphinx_github_changelog",
]

source_suffix = [".rst", ".md"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

myst_enable_extensions = [
    "colon_fence",
]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "furo"


# -- Changelog builder -------------------------------------------------------
if "GITHUB_API_TOKEN" in os.environ:
    sphinx_github_changelog_token = os.environ["GITHUB_API_TOKEN"]

commit = os.environ.get("READTHEDOCS_GIT_COMMIT_HASH", "main")
code_url = "https://github.com/scikit-hep/uhi/blob"



from __future__ import annotations

import copy
import typing
from typing import Any

from .typing.plottable import PlottableAxis

__all__ = ["Locator", "Overflow", "Underflow", "at", "loc", "rebin"]


def __dir__() -> list[str]:
    return __all__


T = typing.TypeVar("T", bound="Locator")


class Locator:
    __slots__ = ("offset",)
    NAME = ""

    def __init__(self, offset: int = 0) -> None:
        if not isinstance(offset, int):
            msg = "The offset must be an integer"  # type: ignore[unreachable]
            raise ValueError(msg)

        self.offset = offset

    def __add__(self: T, offset: int) -> T:
        other = copy.copy(self)
        other.offset += offset
        return other

    def __sub__(self: T, offset: int) -> T:
        other = copy.copy(self)
        other.offset -= offset
        return other

    def _print_self_(self) -> str:
        return ""

    def __repr__(self) -> str:
        s = self.NAME or self.__class__.__name__
        s += self._print_self_()
        if self.offset != 0:
            s += " + " if self.offset > 0 else " - "
            s += str(abs(self.offset))
        return s

    if typing.TYPE_CHECKING:
        # Type checkers think that this is required
        def __index__(self) -> int:
            return 42


class loc(Locator):
    __slots__ = ("value",)

    def __init__(self, value: str | float, offset: int = 0) -> None:
        super().__init__(offset)
        self.value = value

    def _print_self_(self) -> str:
        return f"({self.value})"

    # TODO: clarify that .index() is required
    def __call__(self, axis: Any) -> int:
        return axis.index(self.value) + self.offset  # type: ignore[no-any-return]


class Underflow(Locator):
    __slots__ = ()
    NAME = "underflow"

    def __call__(self, axis: PlottableAxis) -> int:  # noqa: ARG002
        return -1 + self.offset


underflow = Underflow()


class Overflow(Locator):
    __slots__ = ()
    NAME = "overflow"

    def __call__(self, axis: PlottableAxis) -> int:
        return len(axis) + self.offset


overflow = Overflow()


class at:
    __slots__ = ("value",)

    def __init__(self, value: int) -> None:
        self.value = value

    def __call__(self, axis: PlottableAxis) -> int:  # noqa: ARG002
        return self.value


class rebin:
    """
    When used in the step of a Histogram's slice, rebin(n) combines bins,
    scaling their widths by a factor of n. If the number of bins is not
    divisible by n, the remainder is added to the overflow bin.
    """

    def __init__(self, factor: int) -> None:
        # Items with .factor are specially treated in boost-histogram,
        # performing a high performance rebinning
        self.factor = factor

    if typing.TYPE_CHECKING:
        # Type checkers think that this is required
        def __index__(self) -> int:
            return 42



from __future__ import annotations

import numpy as np
import pytest
from pytest import approx

from uhi.numpy_plottable import ensure_plottable_histogram

ROOT = pytest.importorskip("ROOT")


def test_root_imported() -> None:
    assert ROOT.TString("hi") == "hi"


def test_root_th1f_convert() -> None:
    th = ROOT.TH1F("h1", "h1", 50, -2.5, 2.5)
    th.FillRandom("gaus", 10000)
    h = ensure_plottable_histogram(th)
    assert all(th.GetBinContent(i + 1) == approx(iv) for i, iv in enumerate(h.values()))
    var = h.variances()
    assert var is not None
    assert all(th.GetBinError(i + 1) == approx(ie) for i, ie in enumerate(np.sqrt(var)))


def test_root_th2f_convert() -> None:
    th = ROOT.TH2F("h2", "h2", 50, -2.5, 2.5, 50, -2.5, 2.5)
    _ = ROOT.TF2("xyg", "xygaus", -2.5, 2.5, -2.5, 2.5)
    th.FillRandom("xyg", 10000)
    h = ensure_plottable_histogram(th)
    assert all(
        th.GetBinContent(i + 1, j + 1) == approx(iv)
        for i, row in enumerate(h.values())
        for j, iv in enumerate(row)
    )
    var = h.variances()
    assert var is not None
    assert all(
        th.GetBinError(i + 1, j + 1) == approx(ie)
        for i, row in enumerate(np.sqrt(var))
        for j, ie in enumerate(row)
    )



from __future__ import annotations

import contextlib
import importlib.metadata
import sys
import sysconfig
from pathlib import Path

import pytest
from dependency_groups import resolve
from packaging.requirements import Requirement

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

DIR = Path(__file__).parent.resolve()
VALID_FILES = DIR.glob("resources/valid/*.json")
INVALID_FILES = DIR.glob("resources/invalid/*.json")


@pytest.fixture(scope="session")
def resources() -> Path:
    return DIR / "resources"


@pytest.fixture(params=VALID_FILES, ids=lambda p: p.name)
def valid(request: pytest.FixtureRequest) -> Path:
    return request.param  # type: ignore[no-any-return]


@pytest.fixture(params=INVALID_FILES, ids=lambda p: p.name)
def invalid(request: pytest.FixtureRequest) -> Path:
    return request.param  # type: ignore[no-any-return]


@pytest.fixture(params=[False, True], ids=["dense", "sparse"])
def sparse(request: pytest.FixtureRequest) -> bool:
    return request.param  # type: ignore[no-any-return]


def pytest_report_header() -> str:
    with DIR.parent.joinpath("pyproject.toml").open("rb") as f:
        pyproject = tomllib.load(f)
    project = pyproject.get("project", {})

    pkgs = project.get("dependencies", [])
    pkgs += [p for ps in project.get("optional-dependencies", {}).values() for p in ps]

    # Process dependency groups, resolving {include-group = "..."} references
    dep_groups = pyproject.get("dependency-groups", {})
    if dep_groups:
        pkgs += resolve(dep_groups, *dep_groups.keys())

    if "name" in project:
        pkgs.append(project["name"])
    interesting_packages = {Requirement(p).name for p in pkgs}
    interesting_packages.add("pip")
    interesting_packages.add("numpy")

    valid = []
    for package in sorted(interesting_packages):
        with contextlib.suppress(ModuleNotFoundError):
            valid.append(f"{package}=={importlib.metadata.version(package)}")
    reqs = " ".join(valid)
    lines = [
        f"installed packages of interest: {reqs}",
    ]
    if sysconfig.get_config_var("Py_GIL_DISABLED"):
        lines.append("free-threaded Python build")
    return "\n".join(lines)



"""
Using the protocol:

Producers: use isinstance(myhist, PlottableHistogram) in your tests; part of
the protocol is checkable at runtime, though ideally you should use MyPy; if
your histogram class supports PlottableHistogram, this will pass.

Consumers: Make your functions accept the PlottableHistogram static type, and
MyPy will force you to only use items in the Protocol.
"""

from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import Any, Protocol, TypeVar, runtime_checkable

# NumPy 1.20+ will work much, much better than previous versions when type checking
import numpy as np

protocol_version = (1, 2)

# Known kinds of histograms. A Producer can add Kinds not defined here; a
# Consumer should check for known types if it matters. A simple plotter could
# just use .value and .variance if non-None and ignore .kind.
#
# Could have been Kind = Literal["COUNT", "MEAN"] - left as a generic string so
# it can be extendable.
Kind = str

# Implementations are highly encouraged to use the following construct:
# class Kind(str, enum.Enum):
#     COUNT = "COUNT"
#     MEAN = "MEAN"
# Then return and use Kind.COUNT or Kind.MEAN.


@runtime_checkable
class PlottableTraits(Protocol):
    @property
    def circular(self) -> bool:
        """
        True if the axis "wraps around"
        """

    @property
    def discrete(self) -> bool:
        """
        True if each bin is discrete - Integer, Boolean, or Category, for example
        """


T_co = TypeVar("T_co", covariant=True)


@runtime_checkable
class PlottableAxisGeneric(Protocol[T_co]):  # noqa: PLW1641
    # name: str - Optional, not part of Protocol
    # label: str - Optional, not part of Protocol
    #
    # Plotters are encouraged to plot label if it exists and is not None, and
    # name otherwise if it exists and is not None, but these properties are not
    # available on all histograms and not part of the Protocol.

    @property
    def traits(self) -> PlottableTraits: ...

    def __getitem__(self, index: int) -> T_co:
        """
        Get the pair of edges (not discrete) or bin label (discrete).
        """

    def __len__(self) -> int:
        """
        Return the number of bins (not counting flow bins, which are ignored
        for this Protocol currently).
        """

    def __eq__(self, other: Any) -> bool:
        """
        Required to be sequence-like.
        """

    def __iter__(self) -> Iterator[T_co]:
        """
        Useful element of a Sequence to include.
        """


PlottableAxisContinuous = PlottableAxisGeneric[tuple[float, float]]
PlottableAxisInt = PlottableAxisGeneric[int]
PlottableAxisStr = PlottableAxisGeneric[str]

PlottableAxis = PlottableAxisContinuous | PlottableAxisInt | PlottableAxisStr


@runtime_checkable
class PlottableHistogram(Protocol):
    @property
    def axes(self) -> Sequence[PlottableAxis]: ...

    @property
    def kind(self) -> Kind: ...

    # All methods can have a flow=False argument - not part of this Protocol.
    # If this is included, it should return an array with flow bins added,
    # normal ordering.

    def values(self) -> np.typing.NDArray[np.float64]:
        """
        Returns the accumulated values. The counts for simple histograms, the
        sum of weights for weighted histograms, the mean for profiles, etc.

        If counts is equal to 0, the value in that cell is undefined if
        kind == "MEAN".
        """

    def variances(self) -> np.typing.NDArray[np.float64] | None:
        """
        Returns the estimated variance of the accumulated values. The sum of squared
        weights for weighted histograms, the variance of samples for profiles, etc.
        For an unweighed histogram where kind == "COUNT", this should return the same
        as values if the histogram was not filled with weights, and None otherwise.

        If counts is equal to 1 or less, the variance in that cell is undefined if
        kind == "MEAN".

        If kind == "MEAN", the counts can be used to compute the error on the mean
        as sqrt(variances / counts), this works whether or not the entries are
        weighted if the weight variance was tracked by the implementation.
        """

    def counts(self) -> np.typing.NDArray[np.float64] | None:
        """
        Returns the number of entries in each bin for an unweighted
        histogram or profile and an effective number of entries (defined below)
        for a weighted histogram or profile. An exotic generalized histogram could
        have no sensible .counts, so this is Optional and should be checked by
        Consumers.

        If kind == "MEAN", counts (effective or not) can and should be used to
        determine whether the mean value and its variance should be displayed
        (see documentation of values and variances, respectively). The counts
        should also be used to compute the error on the mean (see documentation
        of variances).

        For a weighted histogram, counts is defined as sum_of_weights ** 2 /
        sum_of_weights_squared. It is equal or less than the number of times
        the bin was filled, the equality holds when all filled weights are equal.
        The larger the spread in weights, the smaller it is, but it is always 0
        if filled 0 times, and 1 if filled once, and more than 1 otherwise.

        A suggested implementation is:

            return np.divide(
                sum_of_weights**2,
                sum_of_weights_squared,
                out=np.zeros_like(sum_of_weights, dtype=np.float64),
                where=sum_of_weights_squared != 0)
        """



from __future__ import annotations

import importlib.metadata
import json
from pathlib import Path
from typing import Any

import packaging.version
import pytest
from helpers import convert_histogram_to_32bit

import uhi.io.json
from uhi.io import to_sparse

h5py = pytest.importorskip("h5py", reason="h5py is not installed")
uhi_io_hdf5 = pytest.importorskip("uhi.io.hdf5")

BHVERSION = packaging.version.Version(importlib.metadata.version("boost_histogram"))
HISTVERSION = packaging.version.Version(importlib.metadata.version("hist"))


def test_valid_json(valid: Path, tmp_path: Path, sparse: bool) -> None:
    data = valid.read_text(encoding="utf-8")
    hists = json.loads(data, object_hook=uhi.io.json.object_hook)
    if sparse:
        hists = {name: to_sparse(hist) for name, hist in hists.items()}

    tmp_file = tmp_path / "test.h5"
    with h5py.File(tmp_file, "w") as h5_file:
        for name, hist in hists.items():
            uhi_io_hdf5.write(h5_file.create_group(name), hist)

    with h5py.File(tmp_file, "r") as h5_file:
        rehists = {name: uhi_io_hdf5.read(h5_file[name]) for name in hists}

    assert hists.keys() == rehists.keys()

    for name in hists:
        hist = hists[name]
        rehist = rehists[name]

        # Check that the JSON representation is the same
        data = json.dumps(hist, default=uhi.io.json.default, sort_keys=True)
        redata = json.dumps(rehist, default=uhi.io.json.default, sort_keys=True)

        redata = redata.replace(" ", "").replace("\n", "")
        data = data.replace(" ", "").replace("\n", "")

        assert redata == data


def test_reg_load(tmp_path: Path, resources: Path) -> None:
    data = resources / "valid/reg.json"
    hists = json.loads(
        data.read_text(encoding="utf-8"), object_hook=uhi.io.json.object_hook
    )

    tmp_file = tmp_path / "test.h5"
    with h5py.File(tmp_file, "w") as h5_file:
        uhi_io_hdf5.write(
            h5_file.create_group("one"), hists["one"], min_compress_elements=0
        )

    with h5py.File(tmp_file, "r") as h5_file:
        one = uhi_io_hdf5.read(h5_file["one"])

    assert one["metadata"] == {"one": True, "two": 2, "three": "three"}

    assert len(one["axes"]) == 1
    assert one["axes"][0]["type"] == "regular"
    assert one["axes"][0]["lower"] == pytest.approx(0)
    assert one["axes"][0]["upper"] == pytest.approx(5)
    assert one["axes"][0]["bins"] == 3
    assert one["axes"][0]["underflow"]
    assert one["axes"][0]["overflow"]
    assert not one["axes"][0]["circular"]

    assert one["storage"]["type"] == "int"
    assert one["storage"]["values"] == pytest.approx([1, 2, 3, 4, 5])


@pytest.mark.skipif(
    packaging.version.Version("1.6.1") > BHVERSION,
    reason="Requires boost-histogram 1.6+",
)
def test_convert_bh(tmp_path: Path) -> None:
    import boost_histogram as bh

    h = bh.Histogram(
        bh.axis.Regular(3, 13, 10, __dict__={"name": "x"}), storage=bh.storage.Weight()
    )
    tmp_file = tmp_path / "test.h5"
    with h5py.File(tmp_file, "w") as h5_file:
        uhi_io_hdf5.write(h5_file.create_group("histogram"), h)

    with h5py.File(tmp_file, "r") as h5_file:
        rehist = uhi_io_hdf5.read(h5_file["histogram"])

    h2 = bh.Histogram(rehist)

    assert h == h2


@pytest.mark.skipif(
    packaging.version.Version("1.7.2") > BHVERSION,
    reason="Requires boost-histogram 1.7.2+ for keep_storage=False support",
)
def test_convert_bh_no_storage(tmp_path: Path) -> None:
    """Test HDF5 serialization with keep_storage=False (structure-only histograms)."""
    import boost_histogram as bh
    import boost_histogram.serialization

    h = bh.Histogram(
        bh.axis.Regular(3, 0, 10, __dict__={"name": "x"}), storage=bh.storage.Weight()
    )
    h.fill([0.1, 0.3, 0.5], weight=[1.5, 2.5, 3.5])

    # Convert to UHI format without storage values
    uhi_dict = boost_histogram.serialization.to_uhi(h, keep_storage=False)

    # Verify storage has type but no values
    assert "storage" in uhi_dict
    assert uhi_dict["storage"]["type"] == "weighted"
    assert "values" not in uhi_dict["storage"]

    # Write to HDF5
    tmp_file = tmp_path / "test.h5"
    with h5py.File(tmp_file, "w") as h5_file:
        uhi_io_hdf5.write(h5_file.create_group("histogram"), uhi_dict)

    # Read back from HDF5
    with h5py.File(tmp_file, "r") as h5_file:
        rehist = uhi_io_hdf5.read(h5_file["histogram"])

    # Should be able to reconstruct a histogram with the same structure
    h2 = bh.Histogram(rehist)
    assert h.axes == h2.axes
    assert isinstance(h2.storage_type(), bh.storage.Weight)


@pytest.mark.skipif(
    packaging.version.Version("1.6.1") > BHVERSION
    or packaging.version.Version("2.9.0") > HISTVERSION,
    reason="Requires boost-histogram 1.6+ / Hist 2.9+",
)
def test_convert_hist(tmp_path: Path) -> None:
    import hist

    h = hist.Hist(
        hist.axis.Regular(10, 0, 1, name="a", label="A"),
        hist.axis.Integer(7, 13, overflow=False, name="b", label="B"),
        storage=hist.storage.Weight(),
        name="h",
        label="H",
    )
    tmp_file = tmp_path / "test.h5"
    with h5py.File(tmp_file, "w") as h5_file:
        uhi_io_hdf5.write(h5_file.create_group("histogram"), h)

    with h5py.File(tmp_file, "r") as h5_file:
        rehist = uhi_io_hdf5.read(h5_file["histogram"])
    h2 = hist.Hist(rehist)
    assert h == h2


@pytest.mark.skipif(
    packaging.version.Version("1.6.1") > BHVERSION,
    reason="Requires boost-histogram 1.6+",
)
@pytest.mark.parametrize(
    "storage_type",
    [
        pytest.param("int", id="int_storage"),
        pytest.param("double", id="double_storage"),
        pytest.param("weighted", id="weighted_storage"),
        pytest.param("mean", id="mean_storage"),
    ],
)
def test_convert_bh_32bit_hdf5(tmp_path: Path, storage_type: str) -> None:
    """Test serialization of 32-bit histograms via HDF5."""
    import boost_histogram as bh

    # Create histogram with appropriate storage type
    axis = bh.axis.Regular(5, 0, 1, __dict__={"name": "x"})
    h: bh.Histogram[Any]

    if storage_type == "int":
        h = bh.Histogram(axis, storage=bh.storage.Int64())
        # Fill with some data
        for i in range(5):
            h.fill([0.1 + i * 0.15] * 10)
    elif storage_type == "double":
        h = bh.Histogram(axis, storage=bh.storage.Double())
        h.fill([0.1, 0.3, 0.5, 0.7, 0.9])
    elif storage_type == "weighted":
        h = bh.Histogram(axis, storage=bh.storage.Weight())
        h.fill([0.1, 0.3, 0.5], weight=[1.5, 2.5, 3.5])
    elif storage_type == "mean":
        h = bh.Histogram(axis, storage=bh.storage.Mean())
        h.fill([0.1, 0.3, 0.5], sample=[10.0, 20.0, 30.0])
    else:
        msg = f"Unknown storage type: {storage_type}"
        raise ValueError(msg)

    # Convert to UHI format
    uhi_dict = h._to_uhi_()

    # Convert to 32-bit
    uhi_32bit = convert_histogram_to_32bit(uhi_dict)

    # Write to HDF5
    tmp_file = tmp_path / "test_32bit.h5"
    with h5py.File(tmp_file, "w") as h5_file:
        uhi_io_hdf5.write(h5_file.create_group("histogram"), uhi_32bit)

    # Read back from HDF5
    with h5py.File(tmp_file, "r") as h5_file:
        rehist_32bit = uhi_io_hdf5.read(h5_file["histogram"])

    # Verify storage type and data integrity
    assert rehist_32bit["storage"]["type"] == storage_type
    assert "values" in rehist_32bit["storage"]

    # Verify JSON representation is consistent
    redata = json.dumps(rehist_32bit, default=uhi.io.json.default, sort_keys=True)
    assert len(redata) > 0



from __future__ import annotations

from typing import Any

import numpy as np

from . import ARRAY_KEYS
from ._common import _convert_input

__all__ = ["default", "object_hook"]


def __dir__() -> list[str]:
    return __all__


def default(obj: Any, /) -> Any:
    if hasattr(obj, "_to_uhi_"):
        return _convert_input(obj)
    if isinstance(obj, np.ndarray):
        return obj.tolist()  # Convert ndarray to list
    msg = f"Object of type {type(obj)} is not JSON serializable"
    raise TypeError(msg)


def object_hook(dct: dict[str, Any], /) -> dict[str, Any]:
    """
    Decode a histogram from a dictionary.
    """
    for item in ARRAY_KEYS & dct.keys():
        if isinstance(dct[item], list):
            dct[item] = np.asarray(dct[item])

    return dct



from __future__ import annotations

import abc
import typing
import unittest
from typing import Any

import numpy as np

import uhi.tag
from uhi.typing.serialization import HistogramIR

T = typing.TypeVar("T", bound=Any)

__all__ = ["Indexing1D", "Indexing3D"]


if typing.TYPE_CHECKING:
    # This is a workaround for the issue with type checkers
    # and generics with custom __getitem__ methods.
    sum = 42


def __dir__() -> list[str]:
    return __all__


class Indexing(abc.ABC, unittest.TestCase):
    """
    This super class provides the basic structure for indexing tests.
    """

    def bin_to_value(self, bin: Any) -> Any:
        """
        Allow downstream classes to handle more complex bin objects.
        """
        return bin

    def value_to_bin(self, value: Any) -> Any:
        """
        Inverse of bin_to_value, for downstream classes with complex bins.
        """
        return value

    def values_to_bins(self, values: Any) -> Any:
        """
        Loops over values and converts them to bins using value_to_bin.
        """
        return [self.value_to_bin(v) for v in values]

    def sum_to_value(self, bin: Any) -> Any:
        """
        Allow downstream classes to handle more summed bin objects. E.g. it can be a projection from a Histo2D to Histo1D.
        """
        return bin

    def assertEqualBinValue(self, bin: Any, value: Any) -> None:
        self.assertEqual(self.bin_to_value(bin), value)

    def assertEqualSum(self, bin: Any, value: Any) -> None:
        self.assertEqual(self.sum_to_value(bin), value)


class Indexing1D(typing.Generic[T], Indexing):
    """
    This test requires a histogram to be created first.

    h is a 1D histogram with 10 bins from 0 to 1. Each bin has 2 more than the
    one before, starting with 0. The overflow bin has 1. The underflow bin has 3.
    You can access the UHI serialized version with `.uhi`.
    """

    h: T
    tag = uhi.tag

    @staticmethod
    def get_uhi() -> HistogramIR:
        return {
            "uhi_schema": 1,
            "axes": [
                {
                    "type": "regular",
                    "lower": 0.0,
                    "upper": 1.0,
                    "bins": 10,
                    "underflow": True,
                    "overflow": True,
                    "circular": False,
                }
            ],
            "storage": {
                "type": "double",
                "values": np.array(
                    [3.0, 0.0, 2.0, 4.0, 6.0, 8.0, 10.0, 12.0, 14.0, 16.0, 18.0, 1.0]
                ),
            },
        }

    @classmethod
    @abc.abstractmethod
    def make_histogram(cls) -> T:
        pass

    @classmethod
    def setUpClass(cls) -> None:
        cls.h = cls.make_histogram()

    def test_access_integer(self) -> None:
        for i in range(10):
            with self.subTest(i=i):
                self.assertEqualBinValue(self.h[i], 2 * i)

        with self.assertRaises(IndexError):
            self.h[10]

        self.assertEqualBinValue(self.h[-1], 18)

    def test_access_integer_flow(self) -> None:
        self.assertEqualBinValue(self.h[self.tag.underflow], 3)
        self.assertEqualBinValue(self.h[self.tag.overflow], 1)

    def test_access_loc(self) -> None:
        self.assertEqualBinValue(self.h[self.tag.loc(0.05)], 0)
        self.assertEqualBinValue(self.h[self.tag.loc(0.15)], 2)
        self.assertEqualBinValue(self.h[self.tag.loc(0.95)], 18)
        self.assertEqualBinValue(self.h[self.tag.loc(-1)], 3)
        self.assertEqualBinValue(self.h[self.tag.loc(2)], 1)

    def test_access_loc_addition(self) -> None:
        self.assertEqualBinValue(self.h[self.tag.loc(0.05) + 1], 2)
        self.assertEqualBinValue(self.h[self.tag.loc(0.55) + 2], 14)
        self.assertEqualBinValue(self.h[self.tag.loc(0.55) - 2], 6)

    def test_slicing_all(self) -> None:
        self.assertEqual(self.h[:], self.h)
        self.assertEqual(self.h[...], self.h)

    def test_slicing_closed(self) -> None:
        h = self.h[2:4]
        self.assertEqualBinValue(h[self.tag.underflow], 5)
        self.assertEqualBinValue(h[0], 4)
        self.assertEqualBinValue(h[1], 6)
        self.assertEqualBinValue(h[self.tag.overflow], 79)

        with self.assertRaises(IndexError):
            h[2]

    def test_slicing_open_upper(self) -> None:
        h = self.h[5:]
        self.assertEqualBinValue(h[self.tag.underflow], 23)
        self.assertEqualBinValue(h[0], 10)
        self.assertEqualBinValue(h[4], 18)
        self.assertEqualBinValue(h[self.tag.overflow], 1)
        with self.assertRaises(IndexError):
            h[5]

    def test_slicing_open_lower(self) -> None:
        h = self.h[:5]
        self.assertEqualBinValue(h[self.tag.underflow], 3)
        self.assertEqualBinValue(h[0], 0)
        self.assertEqualBinValue(h[4], 8)
        self.assertEqualBinValue(h[self.tag.overflow], 71)
        with self.assertRaises(IndexError):
            h[5]

    def test_slicing_loc_closed(self) -> None:
        h = self.h[self.tag.loc(0.2) : self.tag.loc(0.4)]
        self.assertEqualBinValue(h[self.tag.underflow], 5)
        self.assertEqualBinValue(h[0], 4)
        self.assertEqualBinValue(h[1], 6)
        self.assertEqualBinValue(h[self.tag.overflow], 79)

        with self.assertRaises(IndexError):
            h[2]

    def test_slicing_loc_open_upper(self) -> None:
        h = self.h[self.tag.loc(0.5) :]
        self.assertEqualBinValue(h[self.tag.underflow], 23)
        self.assertEqualBinValue(h[0], 10)
        self.assertEqualBinValue(h[4], 18)
        self.assertEqualBinValue(h[self.tag.overflow], 1)
        with self.assertRaises(IndexError):
            h[5]

    def test_slicing_loc_open_lower(self) -> None:
        h = self.h[: self.tag.loc(0.5)]
        self.assertEqualBinValue(h[self.tag.underflow], 3)
        self.assertEqualBinValue(h[0], 0)
        self.assertEqualBinValue(h[4], 8)
        self.assertEqualBinValue(h[self.tag.overflow], 71)
        with self.assertRaises(IndexError):
            h[5]

    def test_slicing_loc_mixed(self) -> None:
        h = self.h[2 : self.tag.loc(0.4) + 1]
        self.assertEqualBinValue(h[self.tag.underflow], 5)
        self.assertEqualBinValue(h[0], 4)
        self.assertEqualBinValue(h[1], 6)
        self.assertEqualBinValue(h[2], 8)
        self.assertEqualBinValue(h[self.tag.overflow], 71)

        with self.assertRaises(IndexError):
            h[3]

    def test_rebinning(self) -> None:
        # Boost-histogram allows the `::` to be skipped.
        h = self.h[:: self.tag.rebin(2)]
        self.assertEqualBinValue(h[0], 2)
        self.assertEqualBinValue(h[1], 10)
        self.assertEqualBinValue(h[2], 18)
        self.assertEqualBinValue(h[4], 34)

        with self.assertRaises(IndexError):
            h[5]

    def test_rebinning_with_endpoints(self) -> None:
        h = self.h[1 : 5 : self.tag.rebin(2)]
        self.assertEqualBinValue(h[0], 6)
        self.assertEqualBinValue(h[1], 14)
        with self.assertRaises(IndexError):
            h[2]

    def test_rebinning_with_endpoints_mixed(self) -> None:
        h = self.h[: self.tag.loc(0.55) : self.tag.rebin(2)]
        self.assertEqualBinValue(h[0], 2)
        self.assertEqualBinValue(h[1], 10)

        with self.assertRaises(IndexError):
            h[2]

    def test_full_integration(self) -> None:
        # boost-histogram allows the `::` to be skipped.
        v = self.h[::sum]
        self.assertEqualSum(v, 94)

    def test_non_flow_integration(self) -> None:
        v = self.h[0:len:sum]  # type: ignore[misc]
        self.assertEqualSum(v, 90)

    def test_ranged_integration(self) -> None:
        v = self.h[2:5:sum]
        self.assertEqualSum(v, 18)

    def test_open_lower_integration(self) -> None:
        v = self.h[:4:sum]
        self.assertEqualSum(v, 15)

    def test_open_upper_integration(self) -> None:
        v = self.h[4::sum]
        self.assertEqualSum(v, 79)

    def test_setting_single_value(self) -> None:
        h = self.make_histogram()
        h[0] = self.value_to_bin(42)
        self.assertEqualBinValue(h[0], 42)
        self.assertEqualBinValue(h[1], 2)
        self.assertEqualBinValue(h[-1], 18)

        h[-1] = self.value_to_bin(99)
        self.assertEqualBinValue(h[-1], 99)
        self.assertEqualBinValue(h[0], 42)
        self.assertEqualBinValue(h[1], 2)

    def test_setting_single_value_loc(self) -> None:
        h = self.make_histogram()
        h[self.tag.loc(0.05)] = self.value_to_bin(42)
        self.assertEqualBinValue(h[0], 42)
        self.assertEqualBinValue(h[1], 2)

    def test_setting_underflow(self) -> None:
        h = self.make_histogram()
        h[self.tag.underflow] = self.value_to_bin(42)
        self.assertEqualBinValue(h[self.tag.underflow], 42)
        self.assertEqualBinValue(h[0], 0)

    def test_setting_overflow(self) -> None:
        h = self.make_histogram()
        h[self.tag.overflow] = self.value_to_bin(42)
        self.assertEqualBinValue(h[self.tag.overflow], 42)
        self.assertEqualBinValue(h[-1], 18)

    def test_setting_array(self) -> None:
        h = self.make_histogram()
        h[1:3] = self.value_to_bin(42)

        # TODO: this is broken, fix!
        # self.assertEqual(h[0], 0)
        self.assertEqualBinValue(h[1], 42)
        self.assertEqualBinValue(h[2], 42)
        # self.assertEqual(h[3], 6)

    def test_setting_array_slice(self) -> None:
        h = self.make_histogram()
        h[1:3] = self.values_to_bins([42, 42])

        self.assertEqualBinValue(h[0], 0)
        self.assertEqualBinValue(h[1], 42)
        self.assertEqualBinValue(h[2], 42)
        self.assertEqualBinValue(h[3], 6)

    def test_setting_array_without_underflow(self) -> None:
        h = self.make_histogram()
        h[:3] = self.values_to_bins([42, 43, 44])
        self.assertEqualBinValue(h[self.tag.underflow], 3)
        self.assertEqualBinValue(h[0], 42)
        self.assertEqualBinValue(h[1], 43)
        self.assertEqualBinValue(h[2], 44)
        self.assertEqualBinValue(h[3], 6)

    def test_setting_array_with_underflow(self) -> None:
        h = self.make_histogram()
        h[:3] = self.values_to_bins([41, 42, 43, 44])
        self.assertEqualBinValue(h[self.tag.underflow], 41)
        self.assertEqualBinValue(h[0], 42)
        self.assertEqualBinValue(h[1], 43)
        self.assertEqualBinValue(h[2], 44)
        self.assertEqualBinValue(h[3], 6)

    def test_setting_array_without_overflow(self) -> None:
        h = self.make_histogram()
        h[7:] = self.values_to_bins([42, 43, 44])
        self.assertEqualBinValue(h[6], 12)
        self.assertEqualBinValue(h[7], 42)
        self.assertEqualBinValue(h[8], 43)
        self.assertEqualBinValue(h[9], 44)
        self.assertEqualBinValue(h[self.tag.overflow], 1)

    def test_setting_array_with_overflow(self) -> None:
        h = self.make_histogram()
        h[7:] = self.values_to_bins([42, 43, 44, 45])
        self.assertEqualBinValue(h[6], 12)
        self.assertEqualBinValue(h[7], 42)
        self.assertEqualBinValue(h[8], 43)
        self.assertEqualBinValue(h[9], 44)
        self.assertEqualBinValue(h[self.tag.overflow], 45)

    def test_setting_whole_array(self) -> None:
        h = self.make_histogram()
        h[:] = self.values_to_bins(range(10))
        self.assertEqualBinValue(h[self.tag.underflow], 3)
        self.assertEqualBinValue(h[0], 0)
        self.assertEqualBinValue(h[1], 1)
        self.assertEqualBinValue(h[2], 2)
        self.assertEqualBinValue(h[3], 3)
        self.assertEqualBinValue(h[4], 4)
        self.assertEqualBinValue(h[5], 5)
        self.assertEqualBinValue(h[6], 6)
        self.assertEqualBinValue(h[7], 7)
        self.assertEqualBinValue(h[8], 8)
        self.assertEqualBinValue(h[9], 9)
        self.assertEqualBinValue(h[self.tag.overflow], 1)

    def test_setting_whole_array_with_flow(self) -> None:
        h = self.make_histogram()
        h[:] = self.values_to_bins(range(12))
        self.assertEqualBinValue(h[self.tag.underflow], 0)
        self.assertEqualBinValue(h[0], 1)
        self.assertEqualBinValue(h[1], 2)
        self.assertEqualBinValue(h[2], 3)
        self.assertEqualBinValue(h[3], 4)
        self.assertEqualBinValue(h[4], 5)
        self.assertEqualBinValue(h[5], 6)
        self.assertEqualBinValue(h[6], 7)
        self.assertEqualBinValue(h[7], 8)
        self.assertEqualBinValue(h[8], 9)
        self.assertEqualBinValue(h[9], 10)
        self.assertEqualBinValue(h[self.tag.overflow], 11)

    def test_setting_len_mismatch(self) -> None:
        h = self.make_histogram()

        with self.assertRaises(ValueError):
            h[:] = self.values_to_bins(range(9))
        with self.assertRaises(ValueError):
            h[:] = self.values_to_bins(range(11))
        with self.assertRaises(ValueError):
            h[:] = self.values_to_bins(range(13))

        with self.assertRaises(ValueError):
            h[1:4] = self.values_to_bins(range(2))
        with self.assertRaises(ValueError):
            h[1:4] = self.values_to_bins(range(4))
        with self.assertRaises(ValueError):
            h[1:4] = self.values_to_bins(range(5))


class Indexing2D(typing.Generic[T], Indexing):
    """
    This test requires histograms to be created first.

    h is a 2D histogram with [2,5] bins. The contents are x+2y, where x and y
    are the bin indices.
    """

    h: T
    tag = uhi.tag

    @staticmethod
    def get_uhi() -> HistogramIR:
        x, y = np.mgrid[0:2, 0:5]
        data = np.pad(x + 2 * y, 1, mode="constant")
        return {
            "uhi_schema": 1,
            "axes": [
                {
                    "type": "regular",
                    "lower": 0.0,
                    "upper": 2.0,
                    "bins": 2,
                    "underflow": True,
                    "overflow": True,
                    "circular": False,
                },
                {
                    "type": "regular",
                    "lower": 0.0,
                    "upper": 5.0,
                    "bins": 5,
                    "underflow": True,
                    "overflow": True,
                    "circular": False,
                },
            ],
            "storage": {
                "type": "double",
                "values": data,
            },
        }

    @staticmethod
    @abc.abstractmethod
    def make_histogram() -> T:
        pass

    @classmethod
    def setUpClass(cls) -> None:
        cls.h = cls.make_histogram()

    def test_access_integer(self) -> None:
        for i in range(2):
            for j in range(5):
                with self.subTest(i=i, j=j):
                    self.assertEqualBinValue(self.h[i, j], i + 2 * j)

        with self.assertRaises(IndexError):
            self.h[2, 0]
        with self.assertRaises(IndexError):
            self.h[0, 5]

        self.assertEqualBinValue(self.h[-1, -1], 9)

    def test_access_loc(self) -> None:
        self.assertEqualBinValue(self.h[self.tag.loc(0.5), self.tag.loc(0.5)], 0)
        self.assertEqualBinValue(self.h[self.tag.loc(0.5), self.tag.loc(1.5)], 2)
        self.assertEqualBinValue(self.h[self.tag.loc(0.5), self.tag.loc(4.5)], 8)

    def test_access_loc_underflow(self) -> None:
        self.assertEqualBinValue(self.h[self.tag.loc(-1), self.tag.loc(0.5)], 0)

    def test_access_loc_mixed(self) -> None:
        self.assertEqualBinValue(self.h[self.tag.loc(0.5), 0], 0)
        self.assertEqualBinValue(self.h[0, self.tag.loc(1.5)], 2)
        self.assertEqualBinValue(self.h[1, self.tag.loc(4.5)], 9)

    def test_slicing_all(self) -> None:
        self.assertEqual(self.h[:, :], self.h)
        self.assertEqual(self.h[...], self.h)
        self.assertEqual(self.h[:, ...], self.h)
        self.assertEqual(self.h[..., :], self.h)

    def test_slicing_loc(self) -> None:
        h = self.h[
            1,
            self.tag.loc(0.5) : self.tag.loc(2.5),
        ]
        self.assertEqualBinValue(h[0], 1)
        self.assertEqualBinValue(h[1], 3)

        with self.assertRaises(IndexError):
            h[2]

    def test_rebinning(self) -> None:
        h = self.h[:: self.tag.rebin(2), :: self.tag.rebin(2)]
        self.assertEqualBinValue(h[0, 0], 6)
        self.assertEqualBinValue(h[0, 1], 22)

        with self.assertRaises(IndexError):
            h[0, 2]
        with self.assertRaises(IndexError):
            h[1, 0]

    def test_full_integration(self) -> None:
        v = self.h[::sum, ::sum]
        self.assertEqualSum(v, 45)

    def test_mixed_integration(self) -> None:
        h = self.h[:2:sum, 1:3]
        self.assertEqualSum(h[0], 5)
        self.assertEqualSum(h[1], 9)

        with self.assertRaises(IndexError):
            h[2]

    def test_mixed_single_integration(self) -> None:
        v = self.h[1, ::sum]
        self.assertEqualSum(v, 25)

    def test_mixed_single_integration_dict(self) -> None:
        v = self.h[{0: 1, 1: np.s_[::sum]}]
        self.assertEqualSum(v, 25)

    def test_ellipsis_integration(self) -> None:
        h = self.h[..., ::sum]
        self.assertEqualSum(h[0], 20)
        self.assertEqualSum(h[1], 25)

        with self.assertRaises(IndexError):
            h[2]

    def test_ellipsis_integration_dict(self) -> None:
        h = self.h[{0: np.s_[::sum]}]
        self.assertEqualSum(h[0], 1)
        self.assertEqualSum(h[1], 5)
        self.assertEqualSum(h[4], 17)

        with self.assertRaises(IndexError):
            h[5]

    def test_setting_single_value(self) -> None:
        h = self.make_histogram()
        h[0, 0] = self.value_to_bin(42)
        self.assertEqualBinValue(h[0, 0], 42)
        self.assertEqualBinValue(h[1, 1], 3)

    def test_setting_underflow(self) -> None:
        h = self.make_histogram()
        h[self.tag.underflow, ...] = self.value_to_bin(42)
        self.assertEqualBinValue(h[self.tag.underflow, 0], 42)

    def test_setting_array(self) -> None:
        h = self.make_histogram()

        h[0:2, 0:2] = np.array(
            [self.values_to_bins([42, 43]), self.values_to_bins([44, 45])]
        )
        self.assertEqualBinValue(h[0, 0], 42)
        self.assertEqualBinValue(h[0, 1], 43)
        self.assertEqualBinValue(h[1, 0], 44)
        self.assertEqualBinValue(h[1, 1], 45)
        self.assertEqualBinValue(h[1, 2], 5)

    def test_setting_array_broadcast(self) -> None:
        h = self.make_histogram()

        h[0:2, 0:2] = np.array([self.values_to_bins([42]), self.values_to_bins([3])])
        self.assertEqualBinValue(h[0, 0], 42)
        self.assertEqualBinValue(h[0, 1], 42)
        self.assertEqualBinValue(h[1, 0], 3)
        self.assertEqualBinValue(h[1, 1], 3)
        self.assertEqualBinValue(h[1, 2], 5)

    def test_setting_dict(self) -> None:
        h = self.make_histogram()
        h[{0: 1, 1: 0}] = self.value_to_bin(42)
        self.assertEqualBinValue(h[1, 0], 42)
        self.assertEqualBinValue(h[0, 1], 2)

    def test_setting_dict_slice(self) -> None:
        h = self.make_histogram()

        h[{0: 1, 1: slice(2, 4)}] = self.values_to_bins(range(42, 44))

        self.assertEqualBinValue(h[1, 1], 3)
        self.assertEqualBinValue(h[1, 2], 42)
        self.assertEqualBinValue(h[1, 3], 43)
        self.assertEqualBinValue(h[1, 4], 9)

    def test_setting_dict_slicer(self) -> None:
        h = self.make_histogram()

        h[{0: 1, 1: np.s_[3:5]}] = self.values_to_bins(range(42, 44))
        self.assertEqualBinValue(h[1, 2], 5)
        self.assertEqualBinValue(h[1, 3], 42)
        self.assertEqualBinValue(h[1, 4], 43)


class Indexing3D(typing.Generic[T], Indexing):
    """
    This test requires histograms to be created first.

    h is a 3D histogram with [2,5,10] bins. The contents are x+2y+3z, where x,
    y, and z are the bin indices.
    """

    h: T
    tag = uhi.tag

    @staticmethod
    def get_uhi() -> HistogramIR:
        x, y, z = np.mgrid[0:2, 0:5, 0:10]
        data = np.pad(x + 2 * y + 3 * z, 1, mode="constant")
        return {
            "uhi_schema": 1,
            "writer_info": {"boost-histogram": {"version": "1.6.1"}},
            "axes": [
                {
                    "type": "regular",
                    "lower": 0.0,
                    "upper": 2.0,
                    "bins": 2,
                    "underflow": True,
                    "overflow": True,
                    "circular": False,
                    "metadata": {},
                },
                {
                    "type": "regular",
                    "lower": 0.0,
                    "upper": 5.0,
                    "bins": 5,
                    "underflow": True,
                    "overflow": True,
                    "circular": False,
                    "metadata": {},
                },
                {
                    "type": "regular",
                    "lower": 0.0,
                    "upper": 10.0,
                    "bins": 10,
                    "underflow": True,
                    "overflow": True,
                    "circular": False,
                    "metadata": {},
                },
            ],
            "storage": {
                "type": "double",
                "values": data,
            },
            "metadata": {"_variance_known": True},
        }

    @classmethod
    @abc.abstractmethod
    def make_histogram(cls) -> T:
        pass

    @classmethod
    def setUpClass(cls) -> None:
        cls.h = cls.make_histogram()

    def test_access_integer(self) -> None:
        for i in range(2):
            for j in range(5):
                for k in range(10):
                    with self.subTest(i=i, j=j, k=k):
                        self.assertEqualBinValue(self.h[i, j, k], i + 2 * j + 3 * k)

        with self.assertRaises(IndexError):
            self.h[2, 0, 0]
        with self.assertRaises(IndexError):
            self.h[0, 5, 0]
        with self.assertRaises(IndexError):
            self.h[0, 0, 10]

        self.assertEqualBinValue(self.h[-1, -1, -1], 36)

    def test_access_loc(self) -> None:
        self.assertEqualBinValue(
            self.h[self.tag.loc(0.5), self.tag.loc(0.5), self.tag.loc(0.5)], 0
        )
        self.assertEqualBinValue(
            self.h[self.tag.loc(0.5), self.tag.loc(1.5), self.tag.loc(0.5)], 2
        )
        self.assertEqualBinValue(
            self.h[self.tag.loc(0.5), self.tag.loc(4.5), self.tag.loc(9.5)], 35
        )
        self.assertEqualBinValue(
            self.h[self.tag.loc(-1), self.tag.loc(0.5), self.tag.loc(0.5)], 0
        )

    def test_access_loc_mixed(self) -> None:
        self.assertEqualBinValue(self.h[self.tag.loc(0.5), 0, self.tag.loc(0.5)], 0)
        self.assertEqualBinValue(self.h[0, self.tag.loc(1.5), self.tag.loc(0.5)], 2)
        self.assertEqualBinValue(self.h[self.tag.loc(0.5), self.tag.loc(4.5), 9], 35)
        self.assertEqualBinValue(self.h[1, self.tag.loc(4.5), 9], 36)

    def test_slicing_all(self) -> None:
        self.assertEqual(self.h[:, :, :], self.h)
        self.assertEqual(self.h[...], self.h)
        self.assertEqual(self.h[:, ...], self.h)
        self.assertEqual(self.h[..., :], self.h)

    def test_slicing_loc(self) -> None:
        h = self.h[
            1,
            3,
            self.tag.loc(0.5) : self.tag.loc(2.5),
        ]
        self.assertEqualBinValue(h[0], 7)
        self.assertEqualBinValue(h[1], 10)

        with self.assertRaises(IndexError):
            h[2]

    def test_rebinning(self) -> None:
        h = self.h[:: self.tag.rebin(2), :: self.tag.rebin(2), :: self.tag.rebin(2)]
        self.assertEqualBinValue(h[0, 0, 0], 24)
        self.assertEqualBinValue(h[0, 0, 1], 72)
        self.assertEqualBinValue(h[0, 1, 0], 56)
        self.assertEqualBinValue(h[0, 1, 4], 248)

        with self.assertRaises(IndexError):
            h[0, 1, 5]
        with self.assertRaises(IndexError):
            h[0, 2, 4]
        with self.assertRaises(IndexError):
            h[1, 1, 4]

    def test_full_integration(self) -> None:
        v = self.h[::sum, ::sum, ::sum]
        self.assertEqualSum(v, 1800)

    def test_mixed_integration(self) -> None:
        h = self.h[::sum, :2:sum, 1:3]
        self.assertEqualSum(h[0], 18)
        self.assertEqualSum(h[1], 30)

        with self.assertRaises(IndexError):
            h[2]

    def test_mixed_single_integration(self) -> None:
        h = self.h[1, ::sum, 1:3]
        self.assertEqualSum(h[0], 40)
        self.assertEqualSum(h[1], 55)

        with self.assertRaises(IndexError):
            h[2]

    def test_mixed_single_integration_dict(self) -> None:
        h = self.h[{0: 1, 1: np.s_[::sum], 2: np.s_[1:3]}]
        self.assertEqualSum(h[0], 40)
        self.assertEqualSum(h[1], 55)

        with self.assertRaises(IndexError):
            h[2]

    def test_ellipsis_integration(self) -> None:
        h = self.h[::sum, ..., ::sum]
        self.assertEqualSum(h[0], 280)
        self.assertEqualSum(h[1], 320)
        self.assertEqualSum(h[4], 440)

        with self.assertRaises(IndexError):
            h[5]

    def test_ellipsis_integration_dict(self) -> None:
        h = self.h[{0: np.s_[::sum], 2: np.s_[::sum]}]
        self.assertEqualSum(h[0], 280)
        self.assertEqualSum(h[1], 320)
        self.assertEqualSum(h[4], 440)

        with self.assertRaises(IndexError):
            h[5]

    def test_setting_single_value(self) -> None:
        h = self.make_histogram()
        h[0, 0, 0] = 42
        self.assertEqualBinValue(h[0, 0, 0], 42)
        self.assertEqualBinValue(h[1, 1, 1], 6)

    def test_setting_underflow(self) -> None:
        h = self.make_histogram()
        h[self.tag.underflow, ...] = 42
        self.assertEqualBinValue(h[self.tag.underflow, 0, 0], 42)

    def test_setting_array(self) -> None:
        h = self.make_histogram()

        h[0:2, 0:2, 0:2] = np.array(
            [
                [self.values_to_bins([42, 43]), self.values_to_bins([44, 45])],
                [self.values_to_bins([46, 47]), self.values_to_bins([48, 49])],
            ]
        )
        self.assertEqualBinValue(h[0, 0, 0], 42)
        self.assertEqualBinValue(h[0, 0, 1], 43)
        self.assertEqualBinValue(h[0, 1, 0], 44)
        self.assertEqualBinValue(h[0, 1, 1], 45)
        self.assertEqualBinValue(h[1, 0, 0], 46)
        self.assertEqualBinValue(h[1, 0, 1], 47)
        self.assertEqualBinValue(h[1, 1, 0], 48)
        self.assertEqualBinValue(h[1, 1, 1], 49)
        self.assertEqualBinValue(h[1, 1, 2], 9)

    def test_setting_array_broadcast(self) -> None:
        h = self.make_histogram()

        h[0:2, 0:2, 0:2] = np.array(
            [
                [self.values_to_bins([42]), self.values_to_bins([3])],
                [self.values_to_bins([46]), self.values_to_bins([4])],
            ]
        )
        self.assertEqualBinValue(h[0, 0, 0], 42)
        self.assertEqualBinValue(h[0, 0, 1], 42)
        self.assertEqualBinValue(h[0, 1, 0], 3)
        self.assertEqualBinValue(h[0, 1, 1], 3)
        self.assertEqualBinValue(h[1, 0, 0], 46)
        self.assertEqualBinValue(h[1, 0, 1], 46)
        self.assertEqualBinValue(h[1, 1, 0], 4)
        self.assertEqualBinValue(h[1, 1, 1], 4)
        self.assertEqualBinValue(h[1, 1, 2], 9)

    def test_setting_dict(self) -> None:
        h = self.make_histogram()
        h[{0: 1, 1: 0, 2: 3}] = self.value_to_bin(3)
        self.assertEqualBinValue(h[1, 0, 3], 3)
        self.assertEqualBinValue(h[1, 0, 4], 13)

    def test_setting_dict_slice(self) -> None:
        h = self.make_histogram()

        h[{0: 1, 1: 0, 2: slice(3, 5)}] = self.values_to_bins(range(42, 44))

        self.assertEqualBinValue(h[1, 0, 2], 7)
        self.assertEqualBinValue(h[1, 0, 3], 42)
        self.assertEqualBinValue(h[1, 0, 4], 43)
        self.assertEqualBinValue(h[1, 0, 5], 16)

    def test_setting_dict_slicer(self) -> None:
        h = self.make_histogram()

        h[{0: 1, 1: 0, 2: np.s_[3:5]}] = self.values_to_bins(range(42, 44))
        self.assertEqualBinValue(h[1, 0, 2], 7)
        self.assertEqualBinValue(h[1, 0, 3], 42)
        self.assertEqualBinValue(h[1, 0, 4], 43)
        self.assertEqualBinValue(h[1, 0, 5], 16)



"""
Common helpers for the different formats.
"""

from __future__ import annotations

import typing

from ..typing.serialization import AnyAxisIR, AnyHistogramIR, ToUHIHistogram

__all__ = ["_check_uhi_schema_version", "_convert_input"]


def _check_uhi_schema_version(uhi_schema: int, /) -> None:
    if uhi_schema != 1:
        msg = "Only uhi_schema=1 supported in this uhi version. Please update uhi."
        raise TypeError(msg)


def _remove_empty_metadata(hist: AnyHistogramIR, /) -> AnyHistogramIR:
    if "metadata" in hist and not hist["metadata"]:
        hist = typing.cast(
            AnyHistogramIR, {k: v for k, v in hist.items() if k != "metadata"}
        )
    for i in range(len(hist["axes"])):
        axis = hist["axes"][i]
        if "metadata" in axis and not axis["metadata"]:
            hist["axes"][i] = typing.cast(
                AnyAxisIR, {k: v for k, v in axis.items() if k != "metadata"}
            )
    return hist


def _convert_input(hist: AnyHistogramIR | ToUHIHistogram, /) -> AnyHistogramIR:
    any_hist = typing.cast(
        AnyHistogramIR, hist._to_uhi_() if isinstance(hist, ToUHIHistogram) else hist
    )
    _check_uhi_schema_version(any_hist["uhi_schema"])
    return _remove_empty_metadata(any_hist)



from __future__ import annotations

import dataclasses
from typing import Any

from uhi.io import from_sparse, remove_writer_info, to_sparse
from uhi.io._common import _convert_input
from uhi.typing.serialization import AnyStorageIR


def test_remove_writer_info() -> None:
    d = {"uhi_schema": 1, "writer_info": {"a": {"foo": "bar"}, "b": {"FOO": "BAR"}}}

    assert remove_writer_info(d, library=None) == {"uhi_schema": 1}
    assert remove_writer_info(d, library="a") == {
        "uhi_schema": 1,
        "writer_info": {"b": {"FOO": "BAR"}},
    }
    assert remove_writer_info(d, library="b") == {
        "uhi_schema": 1,
        "writer_info": {"a": {"foo": "bar"}},
    }
    assert remove_writer_info(d, library="c") == d


@dataclasses.dataclass
class _Simple:
    value: dict[str, Any]

    def _to_uhi_(self) -> dict[str, Any]:
        return self.value


def test_remove_empty_metadata() -> None:
    d = {
        "uhi_schema": 1,
        "writer_info": {"boost-histogram": {"version": "1.6.1"}},
        "axes": [
            {
                "type": "regular",
                "lower": 0.0,
                "upper": 1.0,
                "bins": 4,
                "underflow": True,
                "overflow": True,
                "circular": False,
                "metadata": {},
            }
        ],
        "storage": {
            "type": "weighted_mean",
            "sum_of_weights": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            "sum_of_weights_squared": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            "values": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            "variances": [float("nan")] * 6,
        },
        "metadata": {},
    }

    h = _Simple(d)
    ir = _convert_input(h)

    assert "metadata" not in ir
    assert "metadata" not in ir["axes"][0]


def test_empty_storage_conversion() -> None:
    """Test that empty storages are handled correctly during conversion."""
    # Test each storage type with empty arrays
    storage_types: list[tuple[str, dict[str, Any]]] = [
        ("int", {}),
        ("double", {}),
        ("weighted", {}),
        ("mean", {}),
        ("weighted_mean", {}),
    ]

    for storage_type, arrays in storage_types:
        d = {
            "uhi_schema": 1,
            "axes": [
                {
                    "type": "regular",
                    "lower": 0.0,
                    "upper": 1.0,
                    "bins": 4,
                    "underflow": False,
                    "overflow": False,
                    "circular": False,
                }
            ],
            "storage": {"type": storage_type, **arrays},
        }

        h = _Simple(d)
        ir = _convert_input(h)

        storage: AnyStorageIR = ir["storage"]
        assert storage["type"] == storage_type
        # Verify no array fields are present
        array_keys = {
            "values",
            "variances",
            "counts",
            "sum_of_weights",
            "sum_of_weights_squared",
            "index",
        }
        assert not any(k in storage for k in array_keys)


def test_to_sparse_empty_storage() -> None:
    """Test that to_sparse leaves empty storages alone."""
    d = {
        "uhi_schema": 1,
        "axes": [
            {
                "type": "regular",
                "lower": 0.0,
                "upper": 1.0,
                "bins": 4,
                "underflow": False,
                "overflow": False,
                "circular": False,
            }
        ],
        "storage": {"type": "int"},
    }

    result = to_sparse(d)

    # Should return the same histogram unchanged
    assert result == d
    storage_result: AnyStorageIR = result["storage"]  # type: ignore[assignment]
    assert storage_result["type"] == "int"
    assert "values" not in storage_result


def test_from_sparse_empty_storage() -> None:
    """Test that from_sparse leaves empty storages alone."""
    d = {
        "uhi_schema": 1,
        "axes": [
            {
                "type": "regular",
                "lower": 0.0,
                "upper": 1.0,
                "bins": 4,
                "underflow": False,
                "overflow": False,
                "circular": False,
            }
        ],
        "storage": {"type": "double"},
    }

    result = from_sparse(d)

    # Should return the same histogram unchanged
    assert result == d
    storage_result: AnyStorageIR = result["storage"]  # type: ignore[assignment]
    assert storage_result["type"] == "double"
    assert "values" not in storage_result



"""Helper functions for histogram testing."""

from __future__ import annotations

import sys
import typing
from collections.abc import Mapping
from typing import Any

import numpy as np

if sys.version_info < (3, 11):
    from typing import TypeAlias
else:
    from typing import TypeAlias

from uhi.typing.serialization import AnyHistogramIR, AnyStorageIR

__all__ = ["convert_histogram_to_32bit"]

# Type alias for arrays that can be converted
ArrayLike: TypeAlias = (
    np.ndarray | list[typing.Any] | int | np.integer | float | np.floating | None
)


def _convert_array_to_32bit(arr: ArrayLike) -> ArrayLike:
    """Recursively convert numeric arrays to 32-bit precision.

    Uses pattern matching to handle different array types.
    """
    match arr:
        case np.ndarray():
            if np.issubdtype(arr.dtype, np.floating):
                return arr.astype(np.float32)
            if np.issubdtype(arr.dtype, np.signedinteger):
                return arr.astype(np.int32)
            if np.issubdtype(arr.dtype, np.unsignedinteger):
                return arr.astype(np.uint32)
            return arr
        case list():
            return [_convert_array_to_32bit(item) for item in arr]
        case int(n) if n > 2147483647 or n < -2147483648:
            return np.int32(n)
        case _ as other:
            return other


def convert_histogram_to_32bit(hist: Mapping[str, Any]) -> AnyHistogramIR:
    """Convert a histogram dictionary to use 32-bit values where applicable.

    This processes the storage arrays and converts float64 to float32 and
    int64 to int32 (where appropriate).

    Args:
        hist: A histogram in UHI format

    Returns:
        A histogram with 32-bit precision storage arrays
    """
    hist_copy: AnyHistogramIR = typing.cast(AnyHistogramIR, {**hist})

    if "storage" in hist_copy:
        storage: AnyStorageIR = typing.cast(AnyStorageIR, {**hist_copy["storage"]})
        storage_type = storage.get("type")

        match storage_type:
            case "int":
                # Convert integer values to int32
                if "values" in storage:
                    storage["values"] = typing.cast(
                        np.ndarray, _convert_array_to_32bit(storage["values"])
                    )
                if "index" in storage:
                    storage["index"] = typing.cast(
                        np.ndarray, _convert_array_to_32bit(storage["index"])
                    )

            case "double":
                # Convert float values to float32
                if "values" in storage:
                    storage["values"] = typing.cast(
                        np.ndarray, _convert_array_to_32bit(storage["values"])
                    )
                if "index" in storage:
                    storage["index"] = typing.cast(
                        np.ndarray, _convert_array_to_32bit(storage["index"])
                    )

            case "weighted":
                # Convert float values and variances to float32
                if "values" in storage:
                    storage["values"] = typing.cast(
                        np.ndarray, _convert_array_to_32bit(storage["values"])
                    )
                if "variances" in storage:
                    storage["variances"] = typing.cast(
                        np.ndarray, _convert_array_to_32bit(storage["variances"])
                    )
                if "index" in storage:
                    storage["index"] = typing.cast(
                        np.ndarray, _convert_array_to_32bit(storage["index"])
                    )

            case "mean":
                # Convert counts, values, and variances to float32
                if "counts" in storage:
                    storage["counts"] = typing.cast(
                        np.ndarray, _convert_array_to_32bit(storage["counts"])
                    )
                if "values" in storage:
                    storage["values"] = typing.cast(
                        np.ndarray, _convert_array_to_32bit(storage["values"])
                    )
                if "variances" in storage:
                    storage["variances"] = typing.cast(
                        np.ndarray, _convert_array_to_32bit(storage["variances"])
                    )
                if "index" in storage:
                    storage["index"] = typing.cast(
                        np.ndarray, _convert_array_to_32bit(storage["index"])
                    )

            case "weighted_mean":
                # Convert all float arrays to float32
                if "sum_of_weights" in storage:
                    storage["sum_of_weights"] = typing.cast(
                        np.ndarray,
                        _convert_array_to_32bit(storage["sum_of_weights"]),
                    )
                if "sum_of_weights_squared" in storage:
                    storage["sum_of_weights_squared"] = typing.cast(
                        np.ndarray,
                        _convert_array_to_32bit(storage["sum_of_weights_squared"]),
                    )
                if "values" in storage:
                    storage["values"] = typing.cast(
                        np.ndarray, _convert_array_to_32bit(storage["values"])
                    )
                if "variances" in storage:
                    storage["variances"] = typing.cast(
                        np.ndarray, _convert_array_to_32bit(storage["variances"])
                    )
                if "index" in storage:
                    storage["index"] = typing.cast(
                        np.ndarray, _convert_array_to_32bit(storage["index"])
                    )

        hist_copy["storage"] = storage

    return hist_copy



from __future__ import annotations

import json
import re
from pathlib import Path

import fastjsonschema
import pytest

import uhi.schema


def test_valid_schemas(valid: Path) -> None:
    with valid.open(encoding="utf-8") as f:
        data = json.load(f)
    uhi.schema.validate(data)


def test_invalid_schemas(invalid: Path) -> None:
    with invalid.open(encoding="utf-8") as f:
        data = json.load(f)

    try:
        errmsg = invalid.with_suffix(".error.txt").read_text(encoding="utf-8").strip()
    except FileNotFoundError:
        errmsg = "NO ERROR MESSAGE FILE FOUND"

    with pytest.raises(
        fastjsonschema.exceptions.JsonSchemaException, match=re.escape(errmsg)
    ):
        uhi.schema.validate(data)



from __future__ import annotations

import functools
import json
import zipfile
from typing import Any

import numpy as np

from ..typing.serialization import AnyHistogramIR, ToUHIHistogram
from . import ARRAY_KEYS
from ._common import _check_uhi_schema_version, _convert_input

__all__ = ["read", "write"]


def __dir__() -> list[str]:
    return __all__


def write(
    zip_file: zipfile.ZipFile,
    /,
    name: str,
    histogram: AnyHistogramIR | ToUHIHistogram,
) -> None:
    """
    Write a histogram to a zip file.
    """
    histogram = _convert_input(histogram)
    # Write out numpy arrays to files in the zipfile
    for storage_key in ARRAY_KEYS & histogram["storage"].keys():
        path = f"{name}_storage_{storage_key}.npy"
        with zip_file.open(path, "w") as f:
            np.save(f, histogram["storage"][storage_key])  # type: ignore[literal-required]
        histogram["storage"][storage_key] = path  # type: ignore[literal-required]

    for axis in histogram["axes"]:
        for key in ARRAY_KEYS & axis.keys():
            path = f"{name}_axis_{key}.npy"
            with zip_file.open(path, "w") as f:
                np.save(f, axis[key])  # type: ignore[literal-required]
            axis[key] = path  # type: ignore[literal-required]

    hist_json = json.dumps(histogram)
    zip_file.writestr(f"{name}.json", hist_json)


def _object_hook(
    dct: dict[str, Any], /, *, zip_file: zipfile.ZipFile
) -> dict[str, Any]:
    for item in ARRAY_KEYS & dct.keys():
        if isinstance(dct[item], str):
            dct[item] = np.load(zip_file.open(dct[item]))
    return dct


def read(zip_file: zipfile.ZipFile, /, name: str) -> dict[str, Any]:
    """
    Read histograms from a zip file.
    """

    object_hook = functools.partial(_object_hook, zip_file=zip_file)
    with zip_file.open(f"{name}.json") as f:
        output: dict[str, Any] = json.load(f, object_hook=object_hook)
        _check_uhi_schema_version(output["uhi_schema"])
        return output



from __future__ import annotations

import numpy as np
import pytest
from pytest import approx

from uhi.numpy_plottable import ensure_plottable_histogram


def test_from_numpy() -> None:
    hist1 = ((1, 2, 3, 4, 1, 2), (0, 1, 2, 3))

    h = ensure_plottable_histogram(hist1)

    assert h.values() == approx(np.array(hist1[0]))
    assert len(h.axes) == 1
    assert len(h.axes[0]) == 3
    assert h.axes[0][0] == (0, 1)
    assert h.axes[0][1] == (1, 2)
    assert h.axes[0][2] == (2, 3)


def test_from_numpy_2d() -> None:
    np.random.seed(42)
    x = np.random.normal(1, 2, 1000)
    y = np.random.normal(-1, 1, 1000)
    result = np.histogram2d(x, y)

    h = ensure_plottable_histogram(result)

    assert h.values() == approx(result[0])
    assert len(h.axes) == 2
    assert len(h.axes[0]) == 10
    assert h.axes[0][0] == approx(result[1][0:2])
    assert h.axes[0][1] == approx(result[1][1:3])
    assert h.axes[1][0] == approx(result[2][0:2])
    assert h.axes[1][1] == approx(result[2][1:3])


def test_from_numpy_dd() -> None:
    np.random.seed(42)
    x = np.random.normal(1, 2, 1000)
    y = np.random.normal(-1, 1, 1000)
    z = np.random.normal(3, 3, 1000)
    result = np.histogramdd((x, y, z))

    h = ensure_plottable_histogram(result)

    assert h.values() == approx(result[0])
    assert len(h.axes) == 3
    assert len(h.axes[0]) == 10
    assert h.axes[0][0] == approx(result[1][0][0:2])
    assert h.axes[0][1] == approx(result[1][0][1:3])
    assert h.axes[1][0] == approx(result[1][1][0:2])
    assert h.axes[1][1] == approx(result[1][1][1:3])
    assert h.axes[2][0] == approx(result[1][2][0:2])
    assert h.axes[2][1] == approx(result[1][2][1:3])


def test_from_bh_regular() -> None:
    bh = pytest.importorskip("boost_histogram")
    h1 = bh.Histogram(bh.axis.Regular(5, 0, 5))
    h1[...] = (3, 2, 1, 2, 3)

    h = ensure_plottable_histogram(h1)

    assert h is h1

    assert h.values() == approx(np.array((3, 2, 1, 2, 3)))
    assert len(h.axes) == 1
    assert len(h.axes[0]) == 5
    assert h.axes[0][0] == approx(np.array((0, 1)))
    assert h.axes[0][1] == approx(np.array((1, 2)))
    assert h.axes[0][2] == approx(np.array((2, 3)))


def test_from_bh_integer() -> None:
    bh = pytest.importorskip("boost_histogram")
    h1 = bh.Histogram(bh.axis.Integer(1, 6))
    h1[...] = (3, 2, 1, 2, 3)

    h = ensure_plottable_histogram(h1)

    assert h is h1

    assert h.values() == approx(np.array((3, 2, 1, 2, 3)))
    assert len(h.axes) == 1
    assert len(h.axes[0]) == 5
    assert h.axes[0][0] == 1
    assert h.axes[0][1] == 2
    assert h.axes[0][2] == 3


def test_from_numpy_stacked() -> None:
    """Regression test for issue #233.

    A tuple of (list_of_value_arrays, bin_edges) — a stacked histogram format
    used by histoprint — was mishandled: the list was converted to a 2D array
    but only one bin dimension was supplied, causing zip() to raise ValueError.
    """
    rng = np.random.default_rng(42)
    bins = np.linspace(-5, 5, 16)  # 15 bins → 16 edges
    values = [rng.integers(0, 300, size=15) for _ in range(3)]
    hist = (values, bins)

    h = ensure_plottable_histogram(hist)

    assert h.values() == approx(np.asarray(values))
    assert len(h.axes) == 1
    assert len(h.axes[0]) == 15
    assert h.axes[0][0] == approx(bins[0:2])
    assert h.axes[0][-1] == approx(bins[-2:])


def test_from_numpy_excess_bins_raises() -> None:
    """If more bin arrays are provided than histogram dimensions, raise.

    This exercises the new guard that prevents passing e.g. two bin arrays
    for a 1-D values array.
    """
    values = np.arange(5)
    bins1 = np.linspace(0, 5, 6)
    bins2 = np.linspace(0, 5, 6)
    # Provide two bin arrays for a 1-D histogram
    hist = (values, (bins1, bins2))

    with pytest.raises(ValueError, match="Too many bin arrays"):
        ensure_plottable_histogram(hist)


def test_from_bh_str_cat() -> None:
    bh = pytest.importorskip("boost_histogram")
    h1 = bh.Histogram(bh.axis.StrCategory(["hi", "ho"]))
    h1.fill(["hi", "hi", "hi", "ho"])

    h = ensure_plottable_histogram(h1)

    assert h is h1

    assert h.values() == approx(np.array((3, 1)))
    assert len(h.axes) == 1
    assert len(h.axes[0]) == 2

    assert h.axes[0][0] == "hi"
    assert h.axes[0][1] == "ho"



from __future__ import annotations

import boost_histogram as bh

import uhi.testing.indexing


class TestAccess1D(uhi.testing.indexing.Indexing1D[bh.Histogram[bh.storage.Double]]):
    @classmethod
    def make_histogram(cls) -> bh.Histogram[bh.storage.Double]:
        return bh.Histogram(dict(cls.get_uhi()))


class TestAccess2D(uhi.testing.indexing.Indexing2D[bh.Histogram[bh.storage.Double]]):
    @classmethod
    def make_histogram(cls) -> bh.Histogram[bh.storage.Double]:
        return bh.Histogram(dict(cls.get_uhi()))


class TestAccess3D(uhi.testing.indexing.Indexing3D[bh.Histogram[bh.storage.Double]]):
    @classmethod
    def make_histogram(cls) -> bh.Histogram[bh.storage.Double]:
        return bh.Histogram(dict(cls.get_uhi()))


class TestAccessBHTag1D(TestAccess1D):
    tag = bh.tag


class TestAccessBHTag2D(TestAccess2D):
    tag = bh.tag


class TestAccessBHTag3D(TestAccess3D):
    tag = bh.tag



"""
This file holds the NumPyPlottableHistogram, meant to adapt any histogram that
does not support the PlottableHistogram Protocol into a NumPy-backed standin
for it, so plotting functions can remain simple and depend on having a
PlottableHistogram regardless of the input. And this comes with an adaptor
function, ensure_plottable_histogram, which will adapt common input types to a
NumPyPlottibleHistogram, and pass a valid PlottibleHistogram through.

Keep in mind, NumPyPlottableHistogram is a minimal PlottableHistogram instance,
and does not provide any further functionality and is not intended to be used
beyond plotting. Please see a full histogram library like boost-histogram or
hist.
"""

from __future__ import annotations

import abc
import enum
import typing
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, Any

import numpy as np

from uhi.typing.plottable import (
    PlottableAxis,
    PlottableAxisGeneric,
    PlottableHistogram,
    PlottableTraits,
)

if TYPE_CHECKING:
    from numpy.typing import ArrayLike
else:
    ArrayLike = Any


class Kind(str, enum.Enum):
    COUNT = "COUNT"
    MEAN = "MEAN"


class Traits:
    __slots__ = ("circular", "discrete")

    def __init__(self, *, circular: bool = False, discrete: bool = False) -> None:
        self.circular = circular
        self.discrete = discrete


if TYPE_CHECKING:
    _traits: PlottableTraits = typing.cast(Traits, None)


class NumPyPlottableAxis:
    def __init__(self, edges: np.typing.NDArray[Any]) -> None:
        """
        The vals should already be an Nx2 ndarray of edges.
        """
        self.traits: PlottableTraits = Traits()
        assert edges.ndim == 2, "Must be 2D array of edges"
        assert edges.shape[1] == 2, "Second dimension must be 2 (lower, upper)"
        self.edges = edges

    def __repr__(self) -> str:
        """
        Just to be nice for debugging. Not required for the Protocol.
        """

        return f"{self.__class__.__name__}({self.edges!r})"

    def __getitem__(self, index: int) -> tuple[float, float]:
        """
        Get the pair of edges (not discrete) or bin label (discrete).
        """

        return tuple(self.edges[index])

    def __len__(self) -> int:
        """
        Return the number of bins (not counting flow bins, which are ignored
        for this Protocol currently).
        """
        return self.edges.shape[0]  # type: ignore[no-any-return]

    def __eq__(self, other: Any) -> bool:
        """
        Needed for the protocol (should be present to be stored in a Sequence).
        """
        return np.allclose(self.edges, other.edges)

    def __iter__(self) -> Iterator[tuple[float, float]]:
        """
        A useful part of the Protocol for easy access by plotters.
        """
        return iter(self[t] for t in range(len(self)))

    __hash__ = None  # type: ignore[assignment]


if TYPE_CHECKING:
    _axis: PlottableAxisGeneric[tuple[float, float]] = typing.cast(
        NumPyPlottableAxis, None
    )


def _bin_helper(shape: int, bins: np.typing.NDArray[Any] | None) -> NumPyPlottableAxis:
    """
    Returns a axis built from the input bins array, which can be None (0 to N),
    2D lower, upper edges), or 1D (N+1 in length).
    """
    if bins is None:
        return NumPyPlottableAxis(
            np.array([np.arange(0, shape), np.arange(1, shape + 1)]).T
        )
    if bins.ndim == 2:
        return NumPyPlottableAxis(bins)
    if bins.ndim == 1:
        return NumPyPlottableAxis(np.array([bins[:-1], bins[1:]]).T)
    msg = "Bins not understood, should be 2d array of min/max edges or 1D array of edges or None"
    raise ValueError(msg)


class NumPyPlottableHistogram:
    def __init__(
        self,
        hist: np.typing.NDArray[Any],
        *bins: (
            np.typing.NDArray[Any] | None | tuple[np.typing.NDArray[Any] | None, ...]
        ),
        variances: np.typing.NDArray[Any] | None = None,
        kind: Kind = Kind.COUNT,
    ) -> None:
        self._values = hist
        self._variances = variances

        if len(bins) == 1 and isinstance(bins[0], tuple):
            (bins,) = bins  # type: ignore[assignment]

        if len(bins) == 0:
            bins = tuple([None] * len(hist.shape))

        raw_bins: tuple[np.typing.NDArray[Any] | None, ...] = bins  # type: ignore[assignment]

        if len(raw_bins) > len(hist.shape):
            msg = f"Too many bin arrays ({len(raw_bins)}) for histogram with {hist.ndim} dimension(s)"
            raise ValueError(msg)

        self.kind = kind
        # zip over only the last len(raw_bins) dimensions, so that a 2-D
        # "stacked" values array (e.g. multiple 1-D histograms sharing one
        # axis) is handled naturally: the leading dimension(s) are treated as
        # a stacking dimension and produce no axis entry.
        self.axes: Sequence[PlottableAxis] = [
            _bin_helper(shape, b)
            for shape, b in zip(hist.shape[-len(raw_bins) :], raw_bins, strict=True)
        ]

    def __repr__(self) -> str:
        """
        Just to be nice for debugging. Not required for the Protocol.
        """

        axes = ", ".join(repr(s) for s in self.axes)
        return f"{self.__class__.__name__}({self._values!r}, <{axes}>)"

    def values(self) -> np.typing.NDArray[np.float64]:
        return self._values

    def counts(self) -> np.typing.NDArray[np.float64]:
        return self._values

    def variances(self) -> np.typing.NDArray[np.float64] | None:
        return self._variances


if TYPE_CHECKING:
    # Verify that the above class is a valid PlottableHistogram
    _: PlottableHistogram = typing.cast(NumPyPlottableHistogram, None)


def _roottarray_asnumpy(
    tarr: Any, shape: tuple[int, ...] | None = None
) -> np.typing.NDArray[Any]:
    llv = tarr.GetArray()
    arr: np.typing.NDArray[Any] = np.frombuffer(
        llv, dtype=llv.typecode, count=tarr.GetSize()
    )
    if shape is not None:
        return np.reshape(arr, shape, order="F")
    return arr


class ROOTAxis(abc.ABC):
    def __init__(self, tAxis: Any) -> None:
        self.tAx = tAxis

    def __len__(self) -> int:
        return self.tAx.GetNbins()  # type: ignore[no-any-return]

    @abc.abstractmethod
    def __getitem__(self, index: int) -> Any:
        pass

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ROOTAxis):
            return NotImplemented
        return len(self) == len(other) and all(
            aEdges == bEdges for aEdges, bEdges in zip(self, other, strict=True)
        )

    @abc.abstractmethod
    def __iter__(self) -> Iterator[tuple[float, float]] | Iterator[str]:
        pass

    @staticmethod
    def create(tAx: Any) -> DiscreteROOTAxis | ContinuousROOTAxis:
        if all(tAx.GetBinLabel(i + 1) for i in range(tAx.GetNbins())):
            return DiscreteROOTAxis(tAx)
        return ContinuousROOTAxis(tAx)

    __hash__ = None  # type: ignore[assignment]


class ContinuousROOTAxis(ROOTAxis):
    @property
    def traits(self) -> PlottableTraits:
        return Traits(circular=False, discrete=False)

    def __getitem__(self, index: int) -> tuple[float, float]:
        return (self.tAx.GetBinLowEdge(index + 1), self.tAx.GetBinUpEdge(index + 1))

    def __iter__(self) -> Iterator[tuple[float, float]]:
        for i in range(len(self)):
            yield self[i]


class DiscreteROOTAxis(ROOTAxis):
    @property
    def traits(self) -> PlottableTraits:
        return Traits(circular=False, discrete=True)

    def __getitem__(self, index: int) -> str:
        return self.tAx.GetBinLabel(index + 1)  # type: ignore[no-any-return]

    def __iter__(self) -> Iterator[str]:
        for i in range(len(self)):
            yield self[i]


class ROOTPlottableHistBase:
    """Common base for ROOT histograms and TProfile"""

    def __init__(self, thist: Any) -> None:
        self.thist: Any = thist
        nDim = thist.GetDimension()
        self._shape: tuple[int, ...] = tuple(
            getattr(thist, f"GetNbins{ax}")() + 2 for ax in "XYZ"[:nDim]
        )
        self.axes: tuple[ContinuousROOTAxis | DiscreteROOTAxis, ...] = tuple(
            ROOTAxis.create(getattr(thist, f"Get{ax}axis")()) for ax in "XYZ"[:nDim]
        )

    @property
    def name(self) -> str:
        return self.thist.GetName()  # type: ignore[no-any-return]


class ROOTPlottableHistogram(ROOTPlottableHistBase):
    def __init__(self, thist: Any) -> None:
        super().__init__(thist)

    @property
    def hasWeights(self) -> bool:
        return bool(self.thist.GetSumw2() and self.thist.GetSumw2N())

    @property
    def kind(self) -> str:
        return Kind.COUNT

    def values(self) -> np.typing.NDArray[Any]:
        return _roottarray_asnumpy(self.thist, shape=self._shape)[
            tuple([slice(1, -1)] * len(self._shape))
        ]

    def variances(self) -> np.typing.NDArray[Any]:
        if self.hasWeights:
            return _roottarray_asnumpy(self.thist.GetSumw2(), shape=self._shape)[
                tuple([slice(1, -1)] * len(self._shape))
            ]
        return self.values()

    def counts(self) -> np.typing.NDArray[Any]:
        if not self.hasWeights:
            return self.values()

        sumw = self.values()
        result = np.zeros_like(sumw, dtype=np.float64)
        np.divide(
            sumw**2,
            self.variances(),
            out=result,
            where=sumw != 0,
        )
        return result


class ROOTPlottableProfile(ROOTPlottableHistBase):
    def __init__(self, thist: Any) -> None:
        super().__init__(thist)

    @property
    def kind(self) -> str:
        return Kind.MEAN

    def values(self) -> np.typing.NDArray[Any]:
        return np.array(
            [self.thist.GetBinContent(i) for i in range(self.thist.GetNcells())]
        ).reshape(self._shape, order="F")[tuple([slice(1, -1)] * len(self._shape))]

    def variances(self) -> np.typing.NDArray[Any]:
        return (
            np.array([self.thist.GetBinError(i) for i in range(self.thist.GetNcells())])
            ** 2
        ).reshape(self._shape, order="F")[tuple([slice(1, -1)] * len(self._shape))]

    def counts(self) -> np.typing.NDArray[Any]:
        sumw = _roottarray_asnumpy(self.thist, shape=self._shape)[
            tuple([slice(1, -1)] * len(self._shape))
        ]
        if not (self.thist.GetSumw2() and self.thist.GetSumw2N()):
            return sumw

        sumw2 = _roottarray_asnumpy(self.thist.GetSumw2(), shape=self._shape)[
            tuple([slice(1, -1)] * len(self._shape))
        ]
        result = np.zeros_like(sumw, dtype=np.float64)
        np.divide(
            sumw**2,
            sumw2,
            out=result,
            where=sumw != 0,
        )
        return result


if TYPE_CHECKING:
    # Verify that the above class is a valid PlottableHistogram
    _axis = typing.cast(ContinuousROOTAxis, None)
    _axis2: PlottableAxisGeneric[str] = typing.cast(DiscreteROOTAxis, None)
    _ = typing.cast(ROOTPlottableHistogram, None)
    _ = typing.cast(ROOTPlottableProfile, None)


def ensure_plottable_histogram(hist: Any) -> PlottableHistogram:
    """
    Ensure a histogram follows the PlottableHistogram Protocol.

    Currently supports adapting the following inputs:
    * .to_numpy() objects
    * .numpy() objects (uproot3/ROOT)
    * A tuple of NumPy style input. If dd style tuple, must contain
      "np.typing.NDArray[Any]"s. It can contain None's instead of values, including just
      a single None for any number of axes.
    """
    if isinstance(hist, PlottableHistogram):
        # Already satisfies the Protocol
        return hist

    if hasattr(hist, "to_numpy"):
        # Generic (possibly Uproot 4)
        _tup1: tuple[np.typing.NDArray[Any], ...] = hist.to_numpy(flow=False)
        return NumPyPlottableHistogram(*_tup1)

    if hasattr(hist, "numpy"):
        # uproot/TH1 - TODO: could support variances
        _tup2: tuple[np.typing.NDArray[Any], ...] = hist.numpy()
        return NumPyPlottableHistogram(*_tup2)

    if isinstance(hist, tuple):
        # NumPy histogram tuple
        if len(hist) < 2:
            msg = "Can't be applied to less than 2D tuple"
            raise TypeError(msg)
        if (
            len(hist) == 2
            and isinstance(hist[1], list | tuple)
            and all(isinstance(h, np.ndarray) for h in hist[1])
        ):
            # histogramdd tuple
            return NumPyPlottableHistogram(
                np.asarray(hist[0]), *(np.asarray(h) for h in hist[1])
            )
        if hist[1] is None:
            return NumPyPlottableHistogram(
                np.asarray(hist[0]), *(None for _ in np.asarray(hist[0]).shape)
            )
        # Standard tuple
        return NumPyPlottableHistogram(*(np.asarray(h) for h in hist))

    if hasattr(hist, "InheritsFrom") and hist.InheritsFrom("TH1"):
        if any(
            hist.InheritsFrom(profCls)
            for profCls in ("TProfile", "TProfile2D", "TProfile3D")
        ):
            return ROOTPlottableProfile(hist)

        return ROOTPlottableHistogram(hist)

    msg = f"Can't be used on this type of object: {hist!r}"
    raise TypeError(msg)



from __future__ import annotations

import copy
import sys
from typing import Any, TypeVar

import numpy as np

from ..typing.serialization import AnyHistogramIR, AxisIR, HistogramIR

if sys.version_info < (3, 11):
    from typing_extensions import assert_never
else:
    from typing import assert_never


__all__ = ["ARRAY_KEYS", "LIST_KEYS", "from_sparse", "remove_writer_info", "to_sparse"]

ARRAY_KEYS = frozenset(
    [
        "index",
        "values",
        "variances",
        "edges",
        "counts",
        "sum_of_weights",
        "sum_of_weights_squared",
    ]
)

LIST_KEYS = frozenset(
    [
        "categories",
    ]
)

T = TypeVar("T", bound="dict[str, Any]")
H = TypeVar("H", bound="dict[str, Any] | HistogramIR | AnyHistogramIR")


def remove_writer_info(obj: T, /, *, library: str | None) -> T:
    """
    Removes all ``writer_info`` for a library from a histogram dict, axes dict,
    or storage dict. Makes copies where required, and the outer dictionary is
    always copied.

    Specify a library name, or ``None`` to remove all.
    """

    obj = copy.copy(obj)
    if library is None:
        obj.pop("writer_info")
    elif library in obj.get("writer_info", {}):
        obj["writer_info"] = copy.copy(obj["writer_info"])
        del obj["writer_info"][library]

    if "axes" in obj:
        obj["axes"] = [remove_writer_info(ax, library=library) for ax in obj["axes"]]
    if "storage" in obj:
        obj["storage"] = remove_writer_info(obj["storage"], library=library)

    return obj


def _compute_axis_length(axis: AxisIR) -> int:
    if axis["type"] == "regular":
        return axis["bins"] + axis["underflow"] + axis["overflow"]
    if axis["type"] == "variable":
        return len(axis["edges"]) - 1 + axis["underflow"] + axis["overflow"]
    if axis["type"] == "category_str" or axis["type"] == "category_int":
        return len(axis["categories"]) + axis["flow"]
    if axis["type"] == "boolean":
        return 2
    assert_never(axis["type"])


def _zerokey(storage_type: str, key: str) -> bool:
    """
    Returns False if the comparison should be NaN instead of zero.
    """

    return storage_type != "weighted_mean" or key != "variances"


def to_sparse(hist: H, /) -> H:
    """
    Convert a dense histogram to a sparse one. Leaves a sparse histogram alone.
    Leaves empty (metadata-only) histograms alone.
    """

    storage = hist["storage"]
    storage_type = storage["type"]

    # Ignore histograms that have 0 dimensions, are already sparse, or are empty
    if "index" in storage or not hist["axes"] or len(storage) == 1:
        return hist

    # Get the arrays inside storage, ignoring "type"
    arrays = {k: np.asarray(v) for k, v in storage.items() if k != "type"}

    # Build mask of nonzero bins across *all* present keys
    mask = np.any(
        [
            arr != 0 if _zerokey(storage_type, k) else ~np.isnan(arr)
            for k, arr in arrays.items()
        ],
        axis=0,
    )

    # Get the flat indices (or unravel them)
    nonzero_indices = np.nonzero(mask)

    # Pack indices into a single (ndim, n_nonzero) array
    index = np.vstack(nonzero_indices)

    # Build sparse storage dict
    sparse_storage = {"type": storage_type, "index": index}
    for k, arr in arrays.items():
        sparse_storage[k] = arr[mask]

    # Return new histogram dict with modified storage
    sparse_hist = copy.copy(hist)
    sparse_hist["storage"] = sparse_storage  # type: ignore[arg-type]
    return sparse_hist


def from_sparse(sparse: H, /) -> H:
    """
    Convert sparse histogram data back to dense format. If the histogram is already
    dense, just return it.
    """

    storage = sparse["storage"]
    storage_type = storage["type"]

    index = storage.get("index")
    if index is None:
        return sparse

    ndim, _n_nonzero = index.shape
    shape = [_compute_axis_length(a) for a in sparse["axes"]]  # type: ignore[arg-type]

    if len(shape) != ndim:
        msg = f"Shape {shape} does not match sparse index dimension {ndim}"
        raise ValueError(msg)

    dense_storage = {"type": storage["type"]}
    for k, arr1d in storage.items():
        arr1dnp = np.asarray(arr1d)
        if k in {"index", "type"}:
            continue

        # Allocate a zeros (or nan) array of the original shape
        full = np.full(
            shape, 0 if _zerokey(storage_type, k) else np.nan, dtype=arr1dnp.dtype
        )

        # Scatter sparse values back into dense array
        full[tuple(index)] = arr1dnp
        dense_storage[k] = full

    retval = copy.copy(sparse)
    retval["storage"] = dense_storage  # type: ignore[arg-type]
    return retval



"""Serialization types for UHI.

Two types of dictionaries are defined here:

1. ``AnyAxis``, ``AnyStorage``, and ``AnyHistogram`` are used for inputs. They represent
   the merger of all possible types.
2. ``Axis``, ``Storage``, and ``histogram`` are used for outputs. These have precise entries
   defined for each Literal type.

There's also a Protocol, `ToUHIHistogram`, for anything that supports conversion.
"""

from __future__ import annotations

import sys
import typing
from collections.abc import Sequence
from typing import Any, Literal, Protocol, TypedDict

import numpy as np
from numpy.typing import NDArray

if sys.version_info < (3, 11):
    from typing_extensions import NotRequired, Required
else:
    from typing import NotRequired, Required

__all__ = [
    "AnyAxisIR",
    "AnyHistogramIR",
    "AnyStorageIR",
    "AxisIR",
    "BooleanAxisIR",
    "CategoryIntAxisIR",
    "CategoryStrAxisIR",
    "DoubleStorageIR",
    "HistogramIR",
    "IntStorageIR",
    "MeanStorageIR",
    "RegularAxisIR",
    "StorageIR",
    "ToUHIHistogram",
    "VariableAxisIR",
    "WeightedMeanStorageIR",
    "WeightedStorageIR",
]

SupportedMetadata = float | str | bool

ArrayLikeFloat = Sequence[float] | NDArray[np.float64]


def __dir__() -> list[str]:
    return __all__


class RegularAxisIR(TypedDict):
    type: Literal["regular"]
    lower: float
    upper: float
    bins: int
    underflow: bool
    overflow: bool
    circular: bool
    metadata: NotRequired[dict[str, SupportedMetadata]]
    writer_info: NotRequired[dict[str, dict[str, SupportedMetadata]]]


class VariableAxisIR(TypedDict):
    type: Literal["variable"]
    edges: ArrayLikeFloat
    underflow: bool
    overflow: bool
    circular: bool
    metadata: NotRequired[dict[str, SupportedMetadata]]
    writer_info: NotRequired[dict[str, dict[str, SupportedMetadata]]]


class CategoryStrAxisIR(TypedDict):
    type: Literal["category_str"]
    categories: list[str]
    flow: bool
    metadata: NotRequired[dict[str, SupportedMetadata]]
    writer_info: NotRequired[dict[str, dict[str, SupportedMetadata]]]


class CategoryIntAxisIR(TypedDict):
    type: Literal["category_int"]
    categories: list[int]
    flow: bool
    metadata: NotRequired[dict[str, SupportedMetadata]]
    writer_info: NotRequired[dict[str, dict[str, SupportedMetadata]]]


class BooleanAxisIR(TypedDict):
    type: Literal["boolean"]
    metadata: NotRequired[dict[str, SupportedMetadata]]
    writer_info: NotRequired[dict[str, dict[str, SupportedMetadata]]]


class IntStorageIR(TypedDict):
    type: Literal["int"]
    values: NotRequired[NDArray[np.float64]]
    index: NotRequired[NDArray[np.float64]]


class DoubleStorageIR(TypedDict):
    type: Literal["double"]
    values: NotRequired[NDArray[np.float64]]
    index: NotRequired[NDArray[np.float64]]


class WeightedStorageIR(TypedDict):
    type: Literal["weighted"]
    values: NotRequired[NDArray[np.float64]]
    variances: NotRequired[NDArray[np.float64]]
    index: NotRequired[NDArray[np.float64]]


class MeanStorageIR(TypedDict):
    type: Literal["mean"]
    counts: NotRequired[NDArray[np.float64]]
    values: NotRequired[NDArray[np.float64]]
    variances: NotRequired[NDArray[np.float64]]
    index: NotRequired[NDArray[np.float64]]


class WeightedMeanStorageIR(TypedDict):
    type: Literal["weighted_mean"]
    sum_of_weights: NotRequired[NDArray[np.float64]]
    sum_of_weights_squared: NotRequired[NDArray[np.float64]]
    values: NotRequired[NDArray[np.float64]]
    variances: NotRequired[NDArray[np.float64]]
    index: NotRequired[NDArray[np.float64]]


StorageIR = (
    IntStorageIR
    | DoubleStorageIR
    | WeightedStorageIR
    | MeanStorageIR
    | WeightedMeanStorageIR
)


AxisIR = (
    RegularAxisIR
    | VariableAxisIR
    | CategoryStrAxisIR
    | CategoryIntAxisIR
    | BooleanAxisIR
)


class AnyStorageIR(TypedDict, total=False):
    type: Required[Literal["int", "double", "weighted", "mean", "weighted_mean"]]
    index: NDArray[np.float64]
    values: NDArray[np.float64]
    variances: NDArray[np.float64]
    sum_of_weights: NDArray[np.float64]
    sum_of_weights_squared: NDArray[np.float64]
    counts: NDArray[np.float64]


class AnyAxisIR(TypedDict, total=False):
    type: Required[
        Literal["regular", "variable", "category_str", "category_int", "boolean"]
    ]
    metadata: dict[str, SupportedMetadata]
    writer_info: dict[str, dict[str, SupportedMetadata]]
    lower: float
    upper: float
    bins: int
    edges: ArrayLikeFloat
    categories: list[str] | list[int]
    underflow: bool
    overflow: bool
    flow: bool
    circular: bool


class HistogramIR(TypedDict):
    uhi_schema: int
    axes: list[AxisIR]
    storage: StorageIR
    metadata: NotRequired[dict[str, SupportedMetadata]]
    writer_info: NotRequired[dict[str, dict[str, SupportedMetadata]]]


class AnyHistogramIR(TypedDict):
    uhi_schema: int
    axes: list[AnyAxisIR]
    storage: AnyStorageIR
    metadata: NotRequired[dict[str, SupportedMetadata]]
    writer_info: NotRequired[dict[str, dict[str, SupportedMetadata]]]


@typing.runtime_checkable
class ToUHIHistogram(Protocol):
    def _to_uhi_(self) -> dict[str, Any]: ...



from __future__ import annotations

import importlib.metadata
import json
from pathlib import Path
from typing import Any

import packaging.version
import pytest
from helpers import convert_histogram_to_32bit

import uhi.io.json

BHVERSION = packaging.version.Version(importlib.metadata.version("boost_histogram"))
HISTVERSION = packaging.version.Version(importlib.metadata.version("hist"))


def test_valid_json(valid: Path) -> None:
    data = valid.read_text(encoding="utf-8")
    hist = json.loads(data, object_hook=uhi.io.json.object_hook)
    redata = json.dumps(hist, default=uhi.io.json.default)

    rehist = json.loads(redata, object_hook=uhi.io.json.object_hook)
    assert redata.replace(" ", "").replace("\n", "") == data.replace(" ", "").replace(
        "\n", ""
    )

    assert hist.keys() == rehist.keys()


def test_reg_load(resources: Path) -> None:
    data = resources / "valid/reg.json"
    hists = json.loads(
        data.read_text(encoding="utf-8"), object_hook=uhi.io.json.object_hook
    )
    one = hists["one"]
    two = hists["two"]

    assert one["metadata"] == {"one": True, "two": 2, "three": "three"}

    assert len(one["axes"]) == 1
    assert one["axes"][0]["type"] == "regular"
    assert one["axes"][0]["lower"] == pytest.approx(0)
    assert one["axes"][0]["upper"] == pytest.approx(5)
    assert one["axes"][0]["bins"] == 3
    assert one["axes"][0]["underflow"]
    assert one["axes"][0]["overflow"]
    assert not one["axes"][0]["circular"]

    assert one["storage"]["type"] == "int"
    assert one["storage"]["values"] == pytest.approx([1, 2, 3, 4, 5])

    assert len(two["axes"]) == 1
    assert two["axes"][0]["type"] == "regular"
    assert two["axes"][0]["lower"] == pytest.approx(0)
    assert two["axes"][0]["upper"] == pytest.approx(5)
    assert two["axes"][0]["bins"] == 5
    assert two["axes"][0]["underflow"]
    assert two["axes"][0]["overflow"]
    assert not two["axes"][0]["circular"]

    assert two["storage"]["type"] == "double"
    assert two["storage"]["values"] == pytest.approx([1, 2, 3, 4, 5, 6, 7])


@pytest.mark.skipif(
    packaging.version.Version("1.6.1") > BHVERSION,
    reason="Requires boost-histogram 1.6+",
)
def test_convert_bh() -> None:
    import boost_histogram as bh

    h = bh.Histogram(
        bh.axis.Regular(3, 13, 10, __dict__={"name": "x"}), storage=bh.storage.Weight()
    )
    redata = json.dumps(h, default=uhi.io.json.default)
    rehist = json.loads(redata, object_hook=uhi.io.json.object_hook)
    h2 = bh.Histogram(rehist)

    assert h == h2


@pytest.mark.skipif(
    packaging.version.Version("1.7.2") > BHVERSION,
    reason="Requires boost-histogram 1.7.2+ for keep_storage=False support",
)
def test_convert_bh_no_storage() -> None:
    """Test serialization with keep_storage=False (structure-only histograms)."""
    import boost_histogram as bh
    import boost_histogram.serialization

    h = bh.Histogram(
        bh.axis.Regular(3, 0, 10, __dict__={"name": "x"}), storage=bh.storage.Weight()
    )
    h.fill([0.1, 0.3, 0.5], weight=[1.5, 2.5, 3.5])

    # Convert to UHI format without storage values
    uhi_dict = boost_histogram.serialization.to_uhi(h, keep_storage=False)

    # Verify storage has type but no values
    assert "storage" in uhi_dict
    assert uhi_dict["storage"]["type"] == "weighted"
    assert "values" not in uhi_dict["storage"]
    assert "variances" not in uhi_dict["storage"]

    # Should be able to serialize to JSON
    redata = json.dumps(uhi_dict, default=uhi.io.json.default)
    rehist = json.loads(redata, object_hook=uhi.io.json.object_hook)

    # Should be able to reconstruct a histogram with the same structure
    h2 = bh.Histogram(rehist)
    assert h.axes == h2.axes
    assert isinstance(h2.storage_type(), bh.storage.Weight)


@pytest.mark.skipif(
    packaging.version.Version("1.6.1") > BHVERSION
    or packaging.version.Version("2.9.0") > HISTVERSION,
    reason="Requires boost-histogram 1.6+ / Hist 2.9+",
)
def test_convert_hist() -> None:
    import hist

    h = hist.Hist(
        hist.axis.Regular(10, 0, 1, name="a", label="A"),
        hist.axis.Integer(7, 13, overflow=False, name="b", label="B"),
        storage=hist.storage.Weight(),
        name="h",
        label="H",
    )
    redata = json.dumps(h, default=uhi.io.json.default)
    rehist = json.loads(redata, object_hook=uhi.io.json.object_hook)
    h2 = hist.Hist(rehist)
    assert h == h2


def test_empty_storage_roundtrip() -> None:
    """Test that empty storages can be read and written with JSON."""
    data = {
        "empty_int": {
            "uhi_schema": 1,
            "axes": [
                {
                    "type": "regular",
                    "lower": 0,
                    "upper": 10,
                    "bins": 5,
                    "underflow": False,
                    "overflow": False,
                    "circular": False,
                }
            ],
            "storage": {"type": "int"},
        }
    }

    # Dump to JSON
    json_str = json.dumps(data, default=uhi.io.json.default)

    # Load back from JSON
    loaded = json.loads(json_str, object_hook=uhi.io.json.object_hook)

    # Verify structure is preserved
    assert loaded == data
    assert loaded["empty_int"]["storage"]["type"] == "int"
    assert "values" not in loaded["empty_int"]["storage"]


@pytest.mark.skipif(
    packaging.version.Version("1.6.1") > BHVERSION,
    reason="Requires boost-histogram 1.6+",
)
@pytest.mark.parametrize(
    "storage_type",
    [
        pytest.param("int", id="int_storage"),
        pytest.param("double", id="double_storage"),
        pytest.param("weighted", id="weighted_storage"),
        pytest.param("mean", id="mean_storage"),
    ],
)
def test_convert_bh_32bit(storage_type: str) -> None:
    """Test serialization of 32-bit histograms via JSON."""
    import boost_histogram as bh

    # Create histogram with appropriate storage type
    axis = bh.axis.Regular(5, 0, 1, __dict__={"name": "x"})
    h: bh.Histogram[Any]

    if storage_type == "int":
        h = bh.Histogram(axis, storage=bh.storage.Int64())
        # Fill with some data
        for i in range(5):
            h.fill([0.1 + i * 0.15] * 10)
    elif storage_type == "double":
        h = bh.Histogram(axis, storage=bh.storage.Double())
        h.fill([0.1, 0.3, 0.5, 0.7, 0.9])
    elif storage_type == "weighted":
        h = bh.Histogram(axis, storage=bh.storage.Weight())
        h.fill([0.1, 0.3, 0.5], weight=[1.5, 2.5, 3.5])
    elif storage_type == "mean":
        h = bh.Histogram(axis, storage=bh.storage.Mean())
        h.fill([0.1, 0.3, 0.5], sample=[10.0, 20.0, 30.0])
    else:
        msg = f"Unknown storage type: {storage_type}"
        raise ValueError(msg)

    # Convert to UHI format
    uhi_dict = h._to_uhi_()

    # Convert to 32-bit
    uhi_32bit = convert_histogram_to_32bit(uhi_dict)

    # Serialize via JSON
    redata = json.dumps(uhi_32bit, default=uhi.io.json.default)

    # Deserialize
    rehist_32bit = json.loads(redata, object_hook=uhi.io.json.object_hook)

    # Verify storage type and data integrity
    assert rehist_32bit["storage"]["type"] == storage_type
    assert "values" in rehist_32bit["storage"]

    # Verify values can be serialized again without error
    redata2 = json.dumps(rehist_32bit, default=uhi.io.json.default)
    assert len(redata2) > 0



from __future__ import annotations

import json
from pathlib import Path

import numpy as np
import pytest

import uhi.io.json
from uhi.io import from_sparse, to_sparse
from uhi.typing.serialization import HistogramIR, WeightedStorageIR


def test_to_from_sparse_roundtrip() -> None:
    # Original dense data
    hist: HistogramIR = {
        "uhi_schema": 1,
        "storage": {
            "type": "weighted",
            "values": np.array([[0, 1, 0], [0, 2, 0]], dtype=float),
            "variances": np.array([[0, 0.1, 0], [0, 0, 0]], dtype=float),
        },
        "axes": [
            {"type": "boolean"},
            {
                "type": "regular",
                "bins": 3,
                "overflow": False,
                "underflow": False,
                "lower": 0,
                "upper": 1,
                "circular": False,
            },
        ],
    }

    # Convert to sparse
    shist = to_sparse(hist)
    sparse: WeightedStorageIR = shist["storage"]  # type: ignore[assignment]

    # Basic checks on sparse structure
    assert "index" in sparse
    index = sparse["index"]
    assert index.shape[0] == 2
    assert index.shape[1] == 2
    # Verify sparse arrays align with mask
    assert np.all(sparse["values"] == np.array([1.0, 2.0]))
    assert np.all(sparse["variances"] == np.array([0.1, 0.0]))

    # Convert back to dense
    dense = from_sparse(shist)

    # Check round-trip reconstruction
    for key, value in hist["storage"].items():
        if key == "type":
            continue
        assert np.allclose(dense["storage"][key], value)  # type: ignore[literal-required,arg-type]


def test_all_valid(valid: Path) -> None:
    data = valid.read_text(encoding="utf-8")
    hists = json.loads(data, object_hook=uhi.io.json.object_hook)
    hists = {
        k: from_sparse(v) if "index" in v.get("storage", {}) else v
        for k, v in hists.items()
    }
    shists = {k: to_sparse(v) for k, v in hists.items()}
    for h in shists.values():
        # Empty storages (metadata-only) won't get an index since there's no data
        has_data = len(h["storage"]) > 1  # More than just "type"
        if h["axes"] and has_data:
            assert "index" in h["storage"]
    dhists = {k: from_sparse(v) for k, v in shists.items()}
    for h in dhists.values():
        assert "index" not in h["storage"]

    assert hists.keys() == dhists.keys()
    for v, dv in zip(hists.values(), dhists.values(), strict=True):
        assert v.keys() == dv.keys()
        assert v["axes"] == dv["axes"]
        assert v["storage"].keys() == dv["storage"].keys()
        # Empty storages won't have values to compare
        if "values" in v["storage"] and "values" in dv["storage"]:
            assert v["storage"]["values"] == pytest.approx(dv["storage"]["values"])
        if v["storage"]["type"] == "weighted_mean":
            v_var = v["storage"].get("variances")
            dv_var = dv["storage"].get("variances")
            if v_var is not None and dv_var is not None:
                assert np.all(np.isnan(v_var) == np.isnan(dv_var))
                assert v_var[~np.isnan(v_var)] == pytest.approx(
                    dv_var[~np.isnan(dv_var)]
                )
            assert v["storage"].get("sum_of_weights") == pytest.approx(
                dv["storage"].get("sum_of_weights")
            )
            assert v["storage"].get("sum_of_weights_squared") == pytest.approx(
                dv["storage"].get("sum_of_weights_squared")
            )

        else:
            assert v["storage"].get("variances") == pytest.approx(
                dv["storage"].get("variances")
            )
            assert v["storage"].get("counts") == pytest.approx(
                dv["storage"].get("counts")
            )


@pytest.mark.parametrize("name", ["mean", "weighted_mean"])
def test_mean(resources: Path, name: str) -> None:
    data = resources.joinpath("valid/mean.json").read_text()
    hist = json.loads(data, object_hook=uhi.io.json.object_hook)[name]
    sparse_hist = to_sparse(hist)

    assert len(sparse_hist["storage"]["values"]) == 2
    assert sparse_hist["storage"]["index"].shape == (1, 2)



#!/usr/bin/env -S uv run -q

# /// script
# dependencies = ["nox>=2025.2.9"]
# ///

from __future__ import annotations

import argparse
from pathlib import Path

import nox

nox.needs_version = ">=2025.2.9"
nox.options.default_venv_backend = "uv|virtualenv"

PYPROJECT = nox.project.load_toml()
ALL_PYTHONS = nox.project.python_versions(PYPROJECT)

DIR = Path(__file__).parent.resolve()


@nox.session
def lint(session):
    """
    Run the linter.
    """
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", *session.posargs)


@nox.session(python=ALL_PYTHONS)
def tests(session):
    """
    Run the unit and regular tests.
    """
    session.install("-e.[schema]", *nox.project.dependency_groups(PYPROJECT, "test"))
    session.run("pytest", *session.posargs)


@nox.session(reuse_venv=True, default=False)
def docs(session: nox.Session) -> None:
    """
    Build the docs. Use "--non-interactive" to avoid serving. Pass "-b linkcheck" to check links.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-b", dest="builder", default="html", help="Build target (default: html)"
    )
    args, posargs = parser.parse_known_args(session.posargs)

    serve = args.builder == "html" and session.interactive
    extra_installs = ["sphinx-autobuild"] if serve else []
    session.install(
        "-e.", *extra_installs, *nox.project.dependency_groups(PYPROJECT, "docs")
    )

    session.chdir("docs")

    shared_args = (
        "-n",  # nitpicky mode
        "-T",  # full tracebacks
        f"-b={args.builder}",
        ".",
        f"_build/{args.builder}",
        *posargs,
    )

    if serve:
        session.run("sphinx-autobuild", "--open-browser", *shared_args)
    else:
        session.run("sphinx-build", "--keep-going", *shared_args)


@nox.session(default=False)
def build(session):
    """
    Build an SDist and wheel.
    """

    if session.venv_backend == "uv":
        session.run("uv", "build")
    else:
        session.install("build")
        session.run("python", "-m", "build")


@nox.session(venv_backend="conda", default=False)
def root_tests(session):
    """
    Test against ROOT.
    """

    session.conda_install("--channel=conda-forge", "ROOT", "pytest", "boost-histogram")
    session.install("-e.")
    session.run("pytest", "tests/test_root.py")


if __name__ == "__main__":
    nox.run()



from __future__ import annotations

import typing
from typing import Any

import h5py
import numpy as np

from ..typing.serialization import (
    AnyAxisIR,
    AnyHistogramIR,
    AnyStorageIR,
    HistogramIR,
    SupportedMetadata,
    ToUHIHistogram,
)
from . import ARRAY_KEYS
from ._common import _check_uhi_schema_version, _convert_input

__all__ = ["read", "write"]


def __dir__() -> list[str]:
    return __all__


def _handle_metadata_writer_info(
    grp: h5py.Group,
    metadata: dict[str, SupportedMetadata] | None,
    writer_info: dict[str, dict[str, SupportedMetadata]] | None,
) -> None:
    # Metadata
    if metadata:
        metadata_grp = grp.create_group("metadata")
        for key, val1 in metadata.items():
            metadata_grp.attrs[key] = val1

    # Writer info
    if writer_info:
        writer_info_grp = grp.create_group("writer_info")
        for key, value in writer_info.items():
            inner_wi_grp = writer_info_grp.create_group(key)
            for k, v in value.items():
                inner_wi_grp.attrs[k] = v


def write(
    grp: h5py.Group,
    /,
    histogram: AnyHistogramIR | ToUHIHistogram,
    *,
    compression: str = "gzip",
    compression_opts: int = 4,
    min_compress_elements: int = 1_000,
) -> None:
    """
    Write a histogram to an HDF5 group. Arrays larger than
    `min_compress_elements` will be compressed; set to 0 to compress all
    arrays. The `compression` and `compression_opts` arguments are passed
    through.
    """
    histogram = _convert_input(histogram)
    # All referenced objects will be stored inside of /{name}/ref_axes
    hist_folder_storage = grp.create_group("ref_axes")

    # UHI version number
    grp.attrs["uhi_schema"] = histogram["uhi_schema"]

    _handle_metadata_writer_info(
        grp, histogram.get("metadata"), histogram.get("writer_info")
    )

    # Axes
    axes_dataset = grp.create_dataset(
        "axes", len(histogram["axes"]), dtype=h5py.special_dtype(ref=h5py.Reference)
    )
    for i, axis in enumerate(histogram["axes"]):
        # Iterating through the axes, calling `create_axes_object` for each of them,
        # creating references to new groups and appending it to the `items` dataset defined above
        ax_group = hist_folder_storage.create_group(f"axis_{i}")
        ax_info = axis.copy()
        ax_edges_raw = ax_info.pop("edges", None)
        ax_edges = np.asarray(ax_edges_raw) if ax_edges_raw is not None else None
        ax_cats: list[int] | list[str] | None = ax_info.pop("categories", None)
        _handle_metadata_writer_info(
            ax_group, ax_info.pop("metadata", None), ax_info.pop("writer_info", None)
        )
        for key, val2 in ax_info.items():
            ax_group.attrs[key] = val2
        if ax_edges is not None:
            if ax_edges.size < min_compress_elements:
                ax_group.create_dataset("edges", data=ax_edges)
            else:
                ax_group.create_dataset(
                    "edges",
                    data=ax_edges,
                    compression=compression,
                    compression_opts=compression_opts,
                )
        if ax_cats is not None:
            if len(ax_cats) < min_compress_elements:
                ax_group.create_dataset("categories", data=ax_cats)
            else:
                ax_group.create_dataset(
                    "categories",
                    data=ax_cats,
                    compression=compression,
                    compression_opts=compression_opts,
                )
        axes_dataset[i] = ax_group.ref

    # Storage
    storage_grp = grp.create_group("storage")
    storage_type = histogram["storage"]["type"]

    storage_grp.attrs["type"] = storage_type

    for key, val3 in histogram["storage"].items():
        if key == "type":
            continue
        npvalue = np.asarray(val3)
        if npvalue.size < min_compress_elements:
            storage_grp.create_dataset(key, data=npvalue)
        else:
            storage_grp.create_dataset(
                key,
                data=npvalue,
                compression=compression,
                compression_opts=compression_opts,
            )


def _convert_item(name: str, item: Any, /) -> Any:
    """
    Convert an HDF5 item to a native Python type.
    """
    if isinstance(item, bytes):
        return item.decode("utf-8")
    if name == "metadata":
        return {k: _convert_item("", v) for k, v in item.items()}
    if name in ARRAY_KEYS:
        return item
    if isinstance(item, np.generic):
        return item.item()
    return item


def _read_metadata_writer_info(
    output: AnyHistogramIR | AnyAxisIR, group: h5py.Group | h5py.Dataset | h5py.Datatype
) -> None:
    if "metadata" in group:
        output["metadata"] = _convert_item("metadata", group["metadata"].attrs)
    if "writer_info" in group:
        output["writer_info"] = {
            k: _convert_item("metadata", v.attrs)
            for k, v in group["writer_info"].items()
        }


def _convert_axes(group: h5py.Group | h5py.Dataset | h5py.Datatype) -> AnyAxisIR:
    """
    Convert an HDF5 axis reference to a dictionary.
    """
    assert isinstance(group, h5py.Group)

    axis = typing.cast(
        AnyAxisIR, {k: _convert_item(k, v) for k, v in group.attrs.items()}
    )
    if "edges" in group:
        edges = group["edges"]
        assert isinstance(edges, h5py.Dataset)
        axis["edges"] = np.asarray(edges)
    if "categories" in group:
        categories = group["categories"]
        assert isinstance(categories, h5py.Dataset)
        axis["categories"] = [_convert_item("", c) for c in categories]

    _read_metadata_writer_info(axis, group)

    return axis


def read(grp: h5py.Group, /) -> HistogramIR:
    """
    Read a histogram from an HDF5 group.
    """
    uhi_schema = _convert_item("", grp.attrs["uhi_schema"])
    _check_uhi_schema_version(uhi_schema)

    axes_grp = grp["axes"]
    axes_ref = grp["ref_axes"]
    assert isinstance(axes_ref, h5py.Group)
    assert isinstance(axes_grp, h5py.Dataset)

    axes = [_convert_axes(axes_ref[unref_axis_ref]) for unref_axis_ref in axes_ref]

    storage_grp = grp["storage"]
    assert isinstance(storage_grp, h5py.Group)
    storage = AnyStorageIR(type=storage_grp.attrs["type"])
    for key in storage_grp:
        storage[key] = np.asarray(storage_grp[key])  # type: ignore[literal-required]

    histogram_dict = AnyHistogramIR(uhi_schema=uhi_schema, axes=axes, storage=storage)
    _read_metadata_writer_info(histogram_dict, grp)

    return histogram_dict  # type: ignore[return-value]
